from sqlite3 import Time
import a2s  # type:ignore[import]
import requests  # type:ignore[import]

from .server import Server


def query(server: Server) -> None:
    """Queries a server, sets its IP, game port, query port, and status"""
    print(f"Querying {server.name}\n{'-' * (len(server.name) + 9)}")

    result = requests.get(
        f"https://api.battlemetrics.com/servers?filter[search]={server.name}&filter[features][2e079b9a-d6f7-11e7-8461-83e84cedb373]=true&filter"
    )
    print(f"Response: {result}")
    try:
        result.raise_for_status()
    except requests.HTTPError as e:
        print(f"Fetching server failed!\n{e}")
        server.status = "?"
        return

    data = _find_relevant_data(server.name, result.json())
    if server.ip is None:
        ip = server.ip = data["attributes"]["ip"]
        print(f"Found server ip: {ip}")

    if server.game_port is None:
        game_port = server.game_port = data["attributes"]["port"]
        print(f"Found server game port: {game_port}")

    if server.query_port is None:
        query_port = server.query_port = data["attributes"]["portQuery"]
        print(f"Found server query port: {query_port}")

    _set_server_day(server)
    print(f"Server day: {server.day}")

    status = server.status = _get_server_status(server.ip, server.query_port)
    print(f"Server status: {status}\n{'-' * (len(server.name) + 9)}")


def _find_relevant_data(server_name: str, data: dict) -> dict:
    """Extracts the relevant data of a server"""
    for server_data in data["data"]:
        if server_name in server_data["attributes"]["name"]:
            return server_data
            
    raise LookupError(f"Could not find {server_name} in data {data['data']}]")


def _get_server_status(ip: str, port: str) -> str:
    """Queries a servers status by ip and port"""
    try:
        data = a2s.info((ip, port), timeout=7)
        if data:
            return "Ok"
        return "Down"
    except TimeoutError:
        return "Down"


def _set_server_day(server: Server) -> None:
    assert server.ip and server.query_port, "Can't set day without IP"
    try:
        data = a2s.rules((server.ip, server.query_port), timeout=7)
    except TimeoutError:
        print("Failed to query server rules. Server didnt respond!")
        server.status == "Down"
        return
    
    server.day = data["DayTime_s"]



from dataclasses import dataclass


@dataclass
class TribeLogMessage:
    """Represents a single message in the tribe log"""

    day: str
    action: str
    content: str

    def __repr__(self):
        return f"{self.day} {self.action} {self.content}"
    

    def __post_init__(self) -> None:
        self.day = f"Day {self.day[4:]}" 



import pyautogui as pg  # type: ignore[import]

from ark.exceptions import ServerNotFoundError
from ark.server.server import Server

from .._ark import Ark


class SessionList(Ark):
    """Represents the main menu Session List in Ark.

    Provides the ability to rejoin a specific server.
    """

    _FILTERS = {"Official Servers": (369, 863), "Favorites": (373, 925)}
    _SESSION_LIST_REGION = (110, 100, 230, 80)
    _TOP_SERVER_REGION = (90, 200, 430, 90)
    _JOINING_FAILED = (712, 354, 486, 76)

    def refresh(self) -> None:
        """Clicks the refresh button"""
        self.click_at(1229, 941, delay=0.5)

    def join_server(self) -> None:
        self.click_at(990, 943, delay=0.5)

    def open(self) -> None:
        while not self.is_open():
            self.click_at(117, 528, delay=0.5)
            self.sleep(2)

    def connect(self, server: Server) -> None:
        """Joins the given server by the servers search name attribute.

        Parameters:
        ----------
        server :class:`Server`:
            The server to join as a Server object.
        """
        try:
            self.search_server(server)
        except ServerNotFoundError:
            print("Failed to find the server after 15min!")

        while self.is_open():
            self.join_server()
            self.sleep(45)

            if not self._server_is_ghosting():
                continue

            self.sleep(300)
            self.connect(server)

    def search_server(self, server: Server) -> None:
        """Searches for the given server, waits for it to pop up.

        Parameters:
        ----------
        server :class:`Server`:
            The server to search for as a Server object.
        """
        self.click_at(616, 143, delay=0.5)
        with pg.hold("ctrl"):
            pg.press("a")
            pg.press("backspace")

        pg.typewrite(server.name, interval=0.01)
        self.press("enter")
        self.sleep(5)

        # attempt to find the server up to 30 times, waiting 30 seconds each time.
        # after each 30 seconds the session list is refreshed
        attempts = 0
        for _ in range(30):
            count = 0
            while not self.server_found():
                self.sleep(1)
                count += 1

                if count > 30:
                    attempts += 1
                    self.refresh()
                    break
            else:
                self.click_at(959, 237, delay=0.5)
                return

        raise ServerNotFoundError(f"Failed to find {server.name}!")

    def is_open(self) -> bool:
        """Checks if the session list menu is open"""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/session_list.png",
                region=self._SESSION_LIST_REGION,
                confidence=0.8,
            )
            is not None
        )

    def server_found(self) -> bool:
        """Checks if the favorited server has been found"""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/server_favorite.png",
                region=self._TOP_SERVER_REGION,
                confidence=0.75,
            )
            is not None
        )

    def _server_is_ghosting(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/joining_failed.png",
                region=self._JOINING_FAILED,
                confidence=0.7,
            )
            is not None
        )



from typing import Literal, final, overload

from ark.exceptions import InventoryNotAccessibleError, NoItemsDepositedError

from ..._helpers import await_event
from ...items import Item
from ..inventories import DedicatedStorageInventory
from .structure import Structure
from ...config import TIMER_FACTOR


@final
class TekDedicatedStorage(Structure):
    """Represents the Tek Storage in Ark.

    Extends the abilities a regular `Structure` provides by adding
    deposit-interaction related methods.
    """

    _TRANSFERRED_REGION = (710, 4, 460, 130)
    _ITEM_ADDED_REGION = (0, 430, 160, 350)

    def __init__(self) -> None:
        super().__init__("Tek Dedicated Storage", "assets/wheels/dedi.png")
        self.inventory: DedicatedStorageInventory = DedicatedStorageInventory()

    @overload
    def deposit(self, items: list[Item], get_amount: Literal[False]) -> None:
        ...

    @overload
    def deposit(self, items: list[Item], get_amount: Literal[True]) -> tuple[Item, int]:
        ...

    def deposit(self, items: list[Item], get_amount: bool) -> tuple[Item, int] | None:
        """Attempts to deposit into a dedi until the 'x items deposited.'
        green message appears up top where x can be any number.

        Parameters:
        -----------
        items :class:`list`:
            A list of items where each item is a possibly deposited item

        Returns:
        -----------
        item :class:`str`: [Optional]
            The name of the item that was being deposited or was last checked

        amount :class:`int`: [Optional]
            The quantity of items that were deposited, 0 if none.

        Raises:
        -----------
        `NoItemsDepositedError` if the expected text did not appear after 30 seconds.
        """
        # wait for text from possibly prior attempts to go away
        while self.deposited_items():
            self.sleep(0.1)

        self._attempt_deposit()
        if not get_amount:
            return None

        # correct wrong items arg type
        if not isinstance(items, list):
            items = [items]

        for item in items:
            if not self._find_item_deposited(item):
                continue

            print(f"{item.name} was deposited...")
            self.sleep(0.3)
            for _ in range(5):
                if amount := self._get_amount_deposited(item):
                    return item, amount
                self.sleep(0.1)
        return item, 0

    def can_be_opened(self) -> bool:
        """Checks if the dedi can be opened by attempting to do so"""
        try:
            self.inventory.open(max_duration=3)
        except InventoryNotAccessibleError:
            return False
        self.close()
        return True

    def is_in_deposit_range(self) -> bool:
        """Returns whether the dedicated storage is currently within the
        depositing range of the player."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/deposit_all.png",
                region=(0, 0, 1920, 1080),
                confidence=0.7,
            )
            is not None
        )

    def deposited_items(self) -> bool:
        """Returns whether an item has been deposited, determined by the green
        'x items transferred' text on top of the screen."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/items_deposited.png",
                region=self._TRANSFERRED_REGION,
                confidence=0.75,
            )
            is not None
        )

    def _attempt_deposit(self) -> None:
        """Presses the 'E' key on a dedi until the deposit text appears.

        Raises a `NoItemsDepositedError` if we were unable to deposit within
        30 seconds.
        """
        attempts = 0
        while not self.deposited_items():
            self.press(self.keybinds.use)
            if await_event(self.deposited_items, max_duration=3):
                return

            attempts += 1
            if attempts >= (10 * TIMER_FACTOR):
                raise NoItemsDepositedError(
                    f"Failed to deposit after {(10 * TIMER_FACTOR) * 3} seconds!"
                )



from typing import Literal, Optional, overload

import cv2 as cv  # type: ignore[import]
import numpy as np
import PIL  # type: ignore[import]
import pyautogui as pg  # type: ignore[import]
import pygetwindow  # type: ignore[import]
from mss import mss, screenshot, tools  # type: ignore[import]
from PIL import Image, ImageOps
from pytesseract import pytesseract as tes  # type: ignore[import]
from screeninfo import get_monitors  # type: ignore[import]

from ._helpers import get_center
from . import config

class ArkWindow:
    """ARK window handle

    Contains the boundaries of the game and the monitor it is running on.
    Scales points, regions and images depending on the games resolution.

    Contains methods to grab screenshots, match templates and check if the game
    is running. If no ark window could be grabbed, it assumes a regular 1920x1080
    ark window running on a 1920x1080 monitor.

    Properties
    ----------
    window :class:`dict`:
        A dictionary containing the games boundaries

    monitor :class:`dict`:
        A dictionary containing the boundaries of the monitor the game is running on

    fullscreen :class:`bool`:
        Whether the game is running in fullscreen or not.
    """

    _CORRECT_Y = 31
    _CORRECT_X = 8

    def __init__(self) -> None:
        self._boundaries = self.get_boundaries()
        self._monitor = self.get_monitor()
        self._fullscreen = self.check_fullscreen()
        tes.tesseract_cmd = config.TESSERACT_PATH

    def __str__(self) -> str:
        return (
            f"ark window: {self._boundaries}\n"
            f"ark Monitor: {self._monitor}\n"
            f"ark fullscreen: {self._fullscreen}"
        )

    @property
    def handle(self) -> int:
        return self._handle._hWnd

    @property
    def boundaries(self) -> dict:
        return self._boundaries

    @property
    def monitor(self) -> dict:
        return self._monitor

    @property
    def fullscreen(self) -> bool:
        return self._fullscreen

    @property
    def center(self) -> tuple[int, int]:
        return (
            self._boundaries["left"] + (self._boundaries["width"] // 2),
            self._boundaries["top"] + (self._boundaries["height"] // 2),
        )

    @overload
    def grab_screen(
        self, region: tuple[int, int, int, int], path: str, convert: bool = True
    ) -> str:
        ...

    @overload
    def grab_screen(
        self, region: tuple[int, int, int, int], path = None, convert: bool = True
    ) -> screenshot.ScreenShot:
        ...

    def grab_screen(
        self,
        region: tuple[int, int, int, int],
        path: Optional[str] = None,
        convert: bool = True,
    ) -> str:
        """Grabs a screenshot of the given region using mss, if a path
        is provided it will be saved at the path and the path will
        be returned for convenience purposes, otherwise it will simply
        return the `ScreenShot` object.

        Parameters:
        ---------
        Region :class:`tuple`:
            The region of the area to screenshot as (x, y, w, h)

        Path :class:`str`:
            The path to save the image at

        convert :class:`bool`:
            Decides if the given region will be converted or not

        Returns:
        ---------
        The specified path, to improve usage possibilites
        """
        with mss() as sct:
            if convert:
                region = self.convert_region(region)
            x, y, w, h = region

            region_dict = {"left": x, "top": y, "width": w, "height": h}
            img = sct.grab(region_dict)

            if path is None:
                return img
            tools.to_png(img.rgb, img.size, output=path)
            return path

    def set_foreground(self) -> None:
        try:
            self._handle.activate()
        except Exception:
            pass

    def get_boundaries(self) -> dict:
        """Grab the ark window using pygetwindow and create the boundaries.
        If it fails to grab a window it will assume a 1920x1080 window.
        """
        try:
            windows = pygetwindow.getWindowsWithTitle("ARK: Survival Evolved")
            self._handle: pygetwindow.Win32Window = windows[0]
            return {
                "left": self._handle.left + self._CORRECT_X if self._handle.left else 0,
                "top": self._handle.top + self._CORRECT_Y if self._handle.top else 0,
                "width": self._handle.width,
                "height": self._handle.height,
            }
        except Exception as e:
            print(
                "WARNING!\n"
                f"Could not grab the ark boundaries!\n{e}\n\n"
                "Assuming a 1920x1080 windowed fullscreen game."
            )
            return {"left": 0, "top": 0, "width": 1920, "height": 1080}

    def get_monitor(self) -> dict:
        """Gets the monitor boundaries of the monitor ark is running on
        and makes sure its the primary monitor

        Returns:
        ---------
        The boundaries of the monitor ARK is running on
        """
        center = self.center

        for m in get_monitors():
            # check x spacing
            if not (m.x < center[0] and center[0] < m.x + m.width):
                continue

            # check y spacing
            if not (m.y < center[1] and center[1] < m.y + m.height):
                continue

            # create a dict of the monitor
            return {
                "left": m.x,
                "top": m.y,
                "width": m.width,
                "height": m.height,
            }

        print(
            "Could not find then monitor ARK is running on!\n"
            "Assuming a 1920x1080 monitor."
        )
        return self._boundaries

    def check_fullscreen(self) -> bool:
        """Checks if ARK is running in fullscreen by checking if the monitor
        dict matches the ARK boundaries.
        """
        return all(
            [
                self._boundaries[boundary] == self._monitor[boundary]
                for boundary in ["left", "top", "width", "height"]
            ]
        )

    def need_boundary_scaling(self):
        """Checks if we need to scale width and height on regions or images"""
        return False

    def update_boundaries(self):
        """Re-initializes the class to update the window"""
        self._boundaries = self.get_boundaries()
        self._monitor = self.get_monitor()
        self._fullscreen = self.check_fullscreen()

    def convert_width(self, width) -> int:
        """Converts the width if it needs to be scaled."""
        if not self.need_boundary_scaling():
            return width

        return self.convert_point(width, 0)[0]

    def convert_height(self, height) -> int:
        """Converts the width if it needs to be scaled."""
        if not self.need_boundary_scaling():
            return height

        return self.convert_point(0, height)[1]

    def convert_point(self, x=None, y=None):
        """Converts the given point to the corresponding point on the ARK window"""
        # Normalize the position using pyautogui
        x, y = pg._normalizeXYArgs(x, y)

        # check for fullscreen converting
        if self._fullscreen:
            return (
                int((x / 1920) * self.monitor["width"]),
                int((y / 1080) * self.monitor["height"]),
            )

        return (
            self.monitor["left"] + x + self._boundaries["left"],
            self.monitor["top"] + y + self._boundaries["top"],
        )

    def convert_region(self, region: tuple):
        """Converts the given region to the corresponding region on the ARK window"""
        x, y, w, h = region

        # check if we can apply native scaling
        if not self.need_boundary_scaling():
            return (*self.convert_point(x, y), w, h)

        return (*self.convert_point(x, y), *self.convert_point(w, h))

    def convert_image(self, image: str) -> Image.Image | str:
        """Converts the given image to an upscaled image of ARKs resolution.

        Parameters:
        ----------
        image: :class:`str`:
            The path of the image to convert

        Returns:
        ----------
        A PIL `Image.Image` or the path if no converting was needed.
        """
        # check if we need to scale at all
        if not self.need_boundary_scaling():
            return image

        # open the image in PIL and use ImageOps to upscale it (maintains aspect ratio)
        raw_image = Image.open(image)
        new_width = self.convert_point(raw_image.width, raw_image.height)
        return ImageOps.contain(raw_image, new_width, PIL.Image.Resampling.LANCZOS)

    def locate_in_image(
        self, template: str, image, confidence: float, grayscale: bool = False
    ):
        """Finds the location of the given image in the given template."""
        return pg.locate(template, image, confidence=confidence, grayscale=grayscale)

    def locate_all_in_image(
        self, template: str, image, confidence: float, grayscale: bool = False
    ):
        """Finds all locations of the given image in the given template"""
        return self.filter_points(
            set(
                pg.locateAll(
                    template,
                    image,
                    confidence=confidence,
                    grayscale=grayscale,
                )
            ),
            min_dist=15,
        )

    @overload
    def locate_template(
        self,
        template: str,
        region: tuple[int, int, int, int],
        confidence: float,
        *,
        grayscale: bool = False,
        convert: bool = True,
        center: Literal[True],
    ) -> tuple[int, int] | None:
        ...

    @overload
    def locate_template(
        self,
        template: str,
        region: tuple[int, int, int, int],
        confidence: float,
        *,
        grayscale: bool = False,
        convert: bool = True,
        center: Literal[False] = False,
    ) -> tuple[int, int, int, int] | None:
        ...

    def locate_template(
        self,
        template: str,
        region: tuple[int, int, int, int],
        confidence: float,
        *,
        grayscale: bool = False,
        convert: bool = True,
        center: bool = False,
    ) -> tuple[int, int, int, int] | tuple[int, int] | None:
        """Returns the locations of an image on the screen.

        Parameters
        ----------
        template :class:`str` | `Image.Image` | `Mat`:
            The template to find

        image :class:`str` | `Image.Image` | `Mat`:
            The image to find the template in

        confidence :class:`float`:
            How restrictive to be in whats considered a match

        grayscale :class:`bool`: [optional]
            Whehether to grayscale the template, default False

        convert :class:`bool`: [optional]
            Whehether to convert the template, default True

        center :class:`bool`: [optional]
            Whehether to get the matches center, default False

        Returns
        -------
        :class:`tuple[int, int]` | `tuple[int, int, int, int]:
            A tuple with the match either as centers or box
        """
        if convert:
            region = self.convert_region(region)
        haystack: np.ndarray = np.asarray(self.grab_screen(region, convert=False))  # type: ignore[arg-type]
        image_rgb = cv.cvtColor(haystack, cv.COLOR_BGR2RGB)
        img = Image.fromarray(image_rgb)
        box = pg.locate(
            self.convert_image(template) if convert else template,
            img,
            confidence=confidence,
            grayscale=grayscale,
        )
        if not box:
            return None

        box = (box[0] + region[0], box[1] + region[1], box[2], box[3])
        return get_center(box) if box and center else box

    def locate_all_template(
        self,
        template: str,
        region: tuple[int, int, int, int],
        confidence: float,
        convert: bool = True,
        grayscale: bool = False,
    ):
        """Finds all locations of the given template on the screen."""

        haystack: np.ndarray = np.asarray(self.grab_screen(region, convert=False))  # type: ignore[arg-type]
        image_rgb = cv.cvtColor(haystack, cv.COLOR_BGR2RGB)
        img = Image.fromarray(image_rgb)
        return self.filter_points(
            set(
                pg.locateAll(
                    self.convert_image(template),
                    img,
                    confidence=confidence,
                    grayscale=grayscale,
                )
            ),
            min_dist=20,
        )

    def filter_points(self, targets, min_dist) -> set:
        """Filters a set of points by min dist from each other.
        This is important because pyautogui may locate the same template
        multiple times on the same position.
        """
        filtered = set()

        while targets:
            eps = targets.pop()
            for point in targets:
                if all(abs(c2 - c1) < min_dist for c2, c1 in zip(eps, point)):
                    break
            else:
                filtered.add(eps)
        return filtered

    def denoise_text(
        self,
        image: str,
        denoise_rgb: tuple,
        variance: int,
        dilate: bool = True,
        upscale: bool = False,
        upscale_by: int = 8,
    ) -> Image.Image:
        """Denoises / Masks the passed image by the given RGB and variance.
        Useful to pre-process images for a tesseract character scan.

        Parameters:
        ----------
        image :class:`str` | `Image`:
            The image to denoise, if passed as string it will be read using cv2

        denoise_rgb :class:`tuple`:
            The rgb to filter out, only pixels with this rgb will remain in the image

        variance :class:`int`:
            The variance allowed for the denoise_rgb

        Returns:
        ----------
        An upscaled, filtered and dilated version of the given Image as Mat

        """
        # check if we need to read the image or convert it
        if isinstance(image, str):
            image = cv.imread(image, 1)
        else:
            image = np.asarray(image)

        image = cv.cvtColor(image, cv.COLOR_RGB2BGR)
        image = cv.cvtColor(image, cv.COLOR_RGB2BGR)

        # set color range (filering the color of the chars here)
        lower_bound = (
            max(0, denoise_rgb[0] - variance),
            max(0, denoise_rgb[1] - variance),
            max(0, denoise_rgb[2] - variance),
        )
        upper_bound = (
            min(255, denoise_rgb[0] + variance),
            min(255, denoise_rgb[1] + variance),
            min(255, denoise_rgb[2] + variance),
        )

        # load the image into pillow to resize it
        img = cv.inRange(image, lower_bound, upper_bound)

        if not dilate:
            return img

        if upscale:
            img = Image.fromarray(img)
            img = img.resize((img.size[0] * upscale_by, img.size[1] * upscale_by), 1)

        matrix_size = 2 if not upscale else 3

        # Taking a matrix of size 5 as the kernel
        kernel = np.ones((matrix_size, matrix_size), np.uint8)
        return cv.dilate(np.asarray(img), kernel, iterations=1)



from dataclasses import dataclass
from typing import Optional


@dataclass
class Server:
    name: str
    status: Optional[str] = None
    day: Optional[str] = None
    ip: Optional[str] = None
    game_port: Optional[str] = None
    query_port: Optional[str] = None



from .bed import Bed
from .chemistry_bench import ChemistryBench
from .dedicated_storage import TekDedicatedStorage
from .industrial_forge import IndustrialForge
from .industrial_grinder import IndustrialGrinder
from .structure import Structure
from .tek_crop_plot import TekCropPlot
from .sleeping_pod import TekSleepingPod



from ..._ark import Ark
from ...exceptions import DinoNotMountedError, InventoryNotAccessibleError, WheelError
from ...interfaces import ActionWheel, Inventory

class Dinosaur(Ark):
    def __init__(self, entity_name, wheel) -> None:
        super().__init__()
        self.name = entity_name
        self.inventory = Inventory(entity_name)
        self.action_wheel = ActionWheel(entity_name, wheel)

    def access(self) -> None:
        """Wraps the inventory `open` function using the action wheel to
        validate whether we are actually in range of it on failure.

        Raises
        ------
        `InventoryNotAccessibleError`
            If the Inventory could not be accessed even though it is in
            range (determined by the action wheel)

        `WheelNotAccessibleError`
            When no wheel could be opened at all after several attempts

        `UnexpectedWheelError`
            When a wheel was opened but is of an unexpected entity
        """
        try:
            self.inventory.open()
        except InventoryNotAccessibleError as e:
            try:
                self.action_wheel.activate()
            except WheelError:
                self.action_wheel.deactivate()
                raise e
            self.action_wheel.deactivate()
            self.inventory.open(max_duration=40)

    def close(self) -> None:
        self.inventory.close()

    def is_mounted(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/stamina_mount.png",
                region=(1880, 53, 31, 44),
                confidence=0.6,
            )
            is not None
        )

    def can_ride(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/ride.png",
                region=(0, 0, 1920, 1080),
                confidence=0.7,
            )
            is not None
        )

    def can_access(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/access_inventory.png",
                region=(0, 0, 1920, 1080),
                confidence=0.7,
            )
            is not None
        )

    def await_mounted(self) -> bool:
        counter = 0
        while not self.is_mounted():
            self.sleep(0.1)
            if self.inventory.is_open():
                raise DinoNotMountedError(
                    f"Failed to mount {self.name}, accessed inventory instead!"
                    "Please ensure a saddle is equipped."
                )
            counter += 1
            if counter > 130:
                return False
        return True

    def await_dismounted(self) -> bool:
        counter = 0
        while self.is_mounted():
            self.sleep(0.1)
            counter += 1
            if counter > 130:
                return False
        return True

    def mount(self) -> None:
        counter = 0
        while not self.is_mounted():
            self.press(self.keybinds.use)
            if self.await_mounted():
                return
            counter += 1
            if counter >= 4:
                raise DinoNotMountedError(f"Failed to mount {self.name} after 60s!")
        self.sleep(0.5)

    def dismount(self) -> None:
        while self.is_mounted():
            self.press(self.keybinds.use)
            if self.await_dismounted():
                return

        self.sleep(0.5)

    def attack(self, button: str) -> None:
        self.click(button)



from typing import Generator

import cv2 as cv  # type: ignore[import]
import numpy as np
from mss.screenshot import ScreenShot  # type: ignore[import]
from PIL import Image  # type: ignore[import]
from pytesseract import pytesseract as tes  # type: ignore[import]

from ..._ark import Ark
from ...exceptions import LogsNotOpenedError
from .._button import Button
from ._config import (CONTENTS_MAPPING, DAYTIME_MAPPING, DENOISE_MAPPING,
                      EVENT_MAPPING, INGORED_TERMS)
from ._message import TribeLogMessage


class TribeLog(Ark):
    """Represents the ark tribe log. Stores all previous logs as a
    list of `TribeLogMessages`.

    Attributes:
    --------------------
    tribe_log :class:`list`:
        A list containing the past 30 tribe log events as `TribeLogMessages`
    """

    LOG_REGION = 1340, 180, 460, 820

    _ONLINE_AXIS = (1132, 315, 111, 720)

    _TOGGLE_ONLINE = Button(
        (1063, 125), (1035, 97, 52, 52), "toggle_online_members.png"
    )

    def __init__(self) -> None:
        super().__init__()
        self._tribe_log: list[TribeLogMessage] = []
        self._online_members: int | None = None

    def __repr__(self) -> str:
        """A representative string of the log message"""
        return "".join(f"{log_message}\n" for log_message in self._tribe_log)
    
    def __iter__(self) -> Generator[TribeLogMessage, None, None]:
        for message in self._tribe_log:
            yield message

    @property
    def online_members(self) -> str:
        if self._online_members is None:
            return "?"
        if self._online_members >= 12:
            return "12+"
        return str(self._online_members)

    def toggle_online_members(self) -> None:
        assert self._TOGGLE_ONLINE.template and self._TOGGLE_ONLINE.region

        if self.window.locate_template(
            self._TOGGLE_ONLINE.template, self._TOGGLE_ONLINE.region, confidence=0.8
        ):
            return
        self.click_at(self._TOGGLE_ONLINE.location)

    def find_tribelog_events(self, img: ScreenShot) -> list[TribeLogMessage]:
        """Runs a scan on the tribelog snapshot to find all 'Day' occurrences, then
        extracts the message and checks for contents. Adds new messages to the tribelog
        and posts them as alert if they are relevant.
        """
        # sort days from top to bottom by y-coordinate so we can get the message frame
        image_array = np.array(img)
        image_rgb = cv.cvtColor(image_array, cv.COLOR_BGR2RGB)
        image = Image.fromarray(image_rgb)

        day_points = self.get_day_occurrences(image)
        days_in_order = sorted([day for day in day_points], key=lambda t: t[1])

        messages: list[TribeLogMessage] = []
        for i, box in enumerate(days_in_order, start=1):
            try:
                # get relevant regions
                day_region = self.grab_day_region(box)
                message_region = self.grab_message_region(box, days_in_order[i])
            except IndexError:
                break
            try:
                day = self.get_daytime(image.crop(day_region))
                if not day:
                    continue
                
                content = self.get_message_contents(image.crop(message_region))
                if not content:
                    continue
                
                if self.day_is_known(day) or self.content_is_irrelevant(content[1]):
                    continue

            except Exception:
                continue

            # new message with relevant contents, create message object and add it
            # to the new messages
            message = TribeLogMessage(day, *content)
            messages.append(message)

        post = len(self._tribe_log) != 0
        self._tribe_log += reversed(messages)
        self.delete_old_logs()
        return list(reversed(messages)) if post else []

    def grab_current_events(self) -> ScreenShot:
        return self.window.grab_screen(self.LOG_REGION)

    def get_online_members(self) -> None:
        self.toggle_online_members()
        online = len(
            self.window.locate_all_template(
                f"{self.PKG_DIR}/assets/interfaces/online.png",
                region=self._ONLINE_AXIS,
                confidence=0.7,
                grayscale=True,
            )
        )
        self._online_members = online

    def is_open(self) -> bool:
        """Checks if the tribelog is open."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/tribe_log.png",
                region=(1300, 70, 230, 85),
                confidence=0.8,
            )
            is not None
        )

    def open(self) -> None:
        """Opens the tribe log. Tries up to 20 times and raises a
        `LogsNotOpenedError` if unsuccessful.
        """
        c = 0
        while not self.is_open():
            self.press(self.keybinds.logs)
            if self.await_open():
                break
            c += 1
            if c > 20:
                raise LogsNotOpenedError("Failed to open logs!")

        # litle buffer in case timer pops or server lags
        self.sleep(2)

    def close(self) -> None:
        """Closes the tribelogs."""
        while self.is_open():
            self.press("esc")
            if self.await_closed():
                return

    def await_open(self) -> bool:
        """Awaits for the logs to be open to be time efficient.

        to do: write a parent `await` function that does this stuff
        """
        c = 0
        while not self.is_open():
            self.sleep(0.1)
            c += 1
            if c > 10:
                return False
        return True

    def await_closed(self) -> bool:
        """Awaits for the logs to be open to be time efficient."""
        c = 0
        while self.is_open():
            self.sleep(0.1)
            c += 1
            if c > 50:
                return False
        return True

    def grab_day_region(self, box) -> tuple[int, int, int, int]:
        """Grab the day regions located in the top left of the image.
        Adds a little padding on the top, left corner.

        NOTE:
        Int typecasing is neccessary because `box` is of type np.int
        which is not compatible with the cropping.
        """
        return (
            int(box[0] - 5),
            int(box[1] - 5),
            box[0] + 170,
            box[1] + 15,
        )

    def grab_message_region(self, box, next_box) -> tuple[int, int, int, int]:
        """Grabs the region of the message to read using the following message
        as delimiter for the y-axis.

        NOTE:
        Int typecasing is neccessary because `box` is of type np.int
        which is not compatible with the cropping.
        """
        return (
            int(box[0] - 5),
            int(box[1] - 5),
            440,
            int(next_box[1] - 2),
        )

    def content_is_irrelevant(self, content: str) -> bool:
        """Checks if the given content is relevant to be posted.

        Parameters:
        ------------
        content :class:`str`:
            The content to check

        Returns:
        ------------
        Whether an ignored term was found in the contents
        """
        return any(term in content for term in INGORED_TERMS)

    def get_day_occurrences(self, img: Image.Image) -> list[tuple]:
        """Retuns a list of all days, each day being a
        tuple containing top, left, widht and height"""

        img = img.crop(box=(0, 0, 50, img.height))

        return self.window.locate_all_in_image(
            f"{self.PKG_DIR}/assets/tribelog/tribelog_day.png", img, confidence=0.8
        )

    def get_daytime(self, image: str | Image.Image | ScreenShot) -> str | None:
        """Gets the daytime in the given image. The image is denoised and common
        mistakes are filtered out, then the day is validated.

        Parameters:
        -----------
        image :class:`str`| `Image.Image` | `ScreenShot`:
            The image to get the daytime of

        Returns:
        -----------
        daytime :class:`str` | `None`:
            The daytime to be seen in the image or `None` if it is invalid / undetermined.
        """
        # prepare the image, denoising upscaling dilating etc...
        prepared = self.window.denoise_text(
            image, denoise_rgb=(180, 180, 180), variance=18, upscale=True, upscale_by=2
        )

        # get tesseract result, whitelisting seems to not be working too well.
        raw_day_string = tes.image_to_string(prepared, config="--psm 6 -l eng")

        # replace the potentially mistaken characters
        for c in DAYTIME_MAPPING:
            raw_day_string = raw_day_string.replace(c, DAYTIME_MAPPING[c])
        day_string = raw_day_string

        try:
            # split the day into the parts we care about
            day = day_string.split(" ")[1].replace(",", "")
            hour, min, sec = (day_string.split(" ")[2].split(":")[i] for i in range(3))

            # check that all values make logical sense
            if any((len(day) > 5, int(hour) > 24, int(min) > 60, int(sec) > 60)):
                return None

        except Exception as e:
            return None

        return day_string.strip()

    def get_message_contents(
        self, image: str | Image.Image | ScreenShot
    ) -> tuple[str, str] | None:
        """Gets the contents of the given tribelog message image.

        Parameters:
        ------------
        image :class:`str`| `Image.Image` | `ScreenShot`:
            The image to get the contents of

        Returns:
        ------------
        contents :class:`tuple`:
            The contents of the image as a tuple of strings containing the action
            such as "Something destroyed!" and the actual contents.
        """
        # grab the rgb we need to use to denoise the image properly
        # if None there is no meaningful contents in the image
        denoise_rgb = self.get_denoise_rgb(image)
        if not denoise_rgb:
            return None

        # prepare the image for tesseract, the purple RGB needs higher variance
        # than the others. Dilating seems to give accurate results in this case.
        prepared_img = self.window.denoise_text(
            image,
            denoise_rgb,
            30 if not denoise_rgb == (208, 3, 211) else 50,
            upscale=True,
            upscale_by=2,
        )

        # get the raw tesseract result, assuming a uniform block of text.
        raw_res: str = tes.image_to_string(prepared_img, config="--psm 6 -l eng")

        # replace the common known mistakes that tend to happen
        for c in CONTENTS_MAPPING:
            raw_res = raw_res.replace(c, CONTENTS_MAPPING[c])
        filtered_res = raw_res.rstrip()

        if EVENT_MAPPING[denoise_rgb] == "Tek Sensor triggered!":
            sensor_event = self.get_sensor_event(image)
            filtered_res = f"'{filtered_res.rstrip()}' triggered by {sensor_event}!"
            event = "Tek Sensor triggered!"

        if "killed" in filtered_res:
            event = "Something killed!"
        elif "destroyed" in filtered_res:
            event = "Something destroyed!"

        return event, filtered_res

    def get_denoise_rgb(
        self, image: str | Image.Image | ScreenShot
    ) -> tuple[int, int, int] | None:
        """Gets the RGB to denoise for in the given image.

        Parameters:
        ------------
        image :class:`str`| `Image.Image` | `ScreenShot`:
            The image to be denoised

        Returns:
        -----------
        denoise_rgb :class:`tuple`:
            The rgb  value to denoise in the image for a good result
        """
        # absolute pain, need to convert to BGR and back for some reason
        if isinstance(image, ScreenShot):
            image = np.array(image)
            image = cv.cvtColor(image, cv.COLOR_RGB2BGR)
            image = cv.cvtColor(image, cv.COLOR_BGR2RGB)

        elif isinstance(image, str):
            image = cv.imread(image, 1)

        else:
            # convert PIL image to np array for template matching
            image = np.array(image)
            image = cv.cvtColor(image, cv.COLOR_BGR2RGB)
        try:
            # filter out auto-decay
            if self.window.locate_in_image(
                f"{self.PKG_DIR}/assets/tribelog/tribelog_auto_decay.png",
                image,
                confidence=0.8,
            ):
                return None

            # find the RGB we need to denoise
            for rgb in DENOISE_MAPPING:
                template = DENOISE_MAPPING[rgb]

                # only 1 template to match
                if not isinstance(template, list):
                    if (
                        self.window.locate_in_image(template, image, confidence=0.8)
                        is not None
                    ):
                        return rgb
                    continue

                # multiple templates to match, check if any match
                if any(
                    self.window.locate_in_image(template, image, confidence=0.8)
                    is not None
                    for template in DENOISE_MAPPING[rgb]
                ):
                    return rgb
            return None

        except Exception as e:
            print(f"Something went wrong!\n{e}")
            return None

    def get_sensor_event(self, image: Image.Image | str) -> str:
        """Matches for different terms that could have triggered a tek sensor,
        returns the corresponding term.

        Parameters:
        ------------
        image :class:`Any image type compatible with pyautogui.locate`

        Returns:
        ------------
        A string representing the tek sensor event in the passed image.
        """
        if self.window.locate_in_image(
            f"{self.PKG_DIR}/assets/tribelog/tribelog_enemy_survivor.png",
            image,
            confidence=0.75,
        ):
            return "an enemy survivor"

        if self.window.locate_in_image(
            f"{self.PKG_DIR}/assets/tribelog/tribelog_enemy_dino.png",
            image,
            confidence=0.75,
        ):
            return "an enemy dinosaur"
        # not determined
        return "something (friendly or undetermined)"

    def day_is_known(self, day: str) -> bool:
        """Checks if the given day has already been recognized.

        Parameters:
        ------------
        day :class:`str`:
            The day to check for

        Returns:
        ------------
        `True` if the day is already in the database else `False`
        """
        # initial log save, no days known yet
        if not self._tribe_log:
            return False

        # typecast both days to integer to use integer operations
        day_int = int(day.split(" ")[1].replace(",", ""))
        most_recent_day = int(self._tribe_log[-1].day.split(" ")[1].replace(",", ""))

        # check if the day to check is smaller or too high compared to our most recent day
        if day_int < most_recent_day or day_int > most_recent_day + 20:
            return True

        # check if any of the saved messages already contain the day
        for message in self._tribe_log:
            if day.strip() == message.day.strip():
                return True
        return False

    def delete_old_logs(self) -> None:
        """Deletes all but the past 30 messages in the tribelogs."""
        if len(self._tribe_log) < 30:
            return

        self._tribe_log = self._tribe_log[-30:]



import time
from typing import Iterable, Literal, Optional, overload

import pyautogui as pg  # type: ignore[import]
import pydirectinput as input  # type: ignore[import]

from ..._ark import Ark
from ..._helpers import await_event, timedout
from ...buffs import BROKEN_BONES, HUNGRY, THIRSTY, Buff
from ...exceptions import PlayerDidntTravelError, PlayerDiedError
from ...interfaces.hud_info import HUDInfo
from ...interfaces.inventories import Inventory, PlayerInventory
from ...interfaces.structures.structure import Structure
from ...items import Item
from .._stats import Stats


class Player(Ark):
    """Represents the player in ark.

    Provides the ability to control the player to emulate actions a real player
    could do, such as turning or walking, checking on the buffs / debuffs or health,
    and using the hotbar.

    Additionally, it providdes a dedicated `PlayeInventory` which can be accessed
    through the players `inventory` attribute.

    Attributes
    ----------
    inventory :class:`PlayerInventory`:
        The players inventory

    hotbar :class:`list[str]`:
        The players hotbar slots
    """
    _DEBUFF_REGION = (1270, 950, 610, 130)
    _ADDED_REGION = (0, 450, 314, 240)
    _HP_BAR = (1882, 1022, 15, 50)
    _HAS_DIED = (630, 10, 590, 80)
    _STAM_BAR = (1850, 955, 70, 65)

    @overload
    def __init__(self, *, stats: Stats) -> None:
        ...

    @overload
    def __init__(self, health: int, food: int, water: int, weight: int) -> None:
        ...

    def __init__(
        self, health=None, food=None, water=None, weight=None, *, stats=None
    ) -> None:
        super().__init__()
        self.inventory = PlayerInventory()
        if stats is not None:
            self.stats = stats
        else:
            self.stats = Stats(health, food, water, weight)
        self.HOTBAR = [
            self.keybinds.hotbar_1,
            self.keybinds.hotbar_2,
            self.keybinds.hotbar_3,
            self.keybinds.hotbar_4,
            self.keybinds.hotbar_5,
            self.keybinds.hotbar_6,
            self.keybinds.hotbar_7,
            self.keybinds.hotbar_8,
            self.keybinds.hotbar_9,
            self.keybinds.hotbar_0,
        ]
        self.hud = HUDInfo()
        self._lr_factor = 3.2 / self.settings.left_right_sens
        self._ud_factor = 3.2 / self.settings.up_down_sens
        self._fov_factor = 1.25 / self.settings.fov_multiplier

    def turn_90_degrees(
        self, direction: Literal["right", "left"] = "right", delay: int | float = 0
    ) -> None:
        """Turns by 90 degrees in given direction"""
        val = 129 if direction == "right" else -129
        self.turn_x_by(val)
        self.sleep(delay)

    def look_down_hard(self) -> None:
        """Looks down all the way to the ground"""
        for _ in range(7):
            self.turn_y_by(50, delay=0.05)
        self.sleep(0.3)

    def look_up_hard(self) -> None:
        """Looks up all the way to the ceiling"""
        for _ in range(7):
            self.turn_y_by(-50, delay=0.05)
        self.sleep(0.3)

    def turn_y_by(self, amount: int, delay: int | float = 0.1) -> None:
        """Turns the players' y-axis by the given amount"""
        input.moveRel(
            0,
            round(amount * self._ud_factor * self._fov_factor),
            0,
            None,
            False,
            False,
            True,
        )
        self.sleep(delay)

    def turn_x_by(self, amount: int, delay: int | float = 0.1) -> None:
        """Turns the players' x-axis by the given amount"""
        input.moveRel(
            round(amount * self._lr_factor * self._fov_factor),
            0,
            0,
            None,
            False,
            False,
            True,
        )
        self.sleep(delay)

    def pick_up(self) -> None:
        """Picks up an item by pressing E"""
        self.press(self.keybinds.use)

    def pick_all(self) -> None:
        """Picks all items by pressing F"""
        self.press(self.keybinds.target_inventory)

    def drop_all(self, items: Optional[Iterable[Item | str]] = None) -> None:
        """Opens the inventory and drops all on the specified item"""
        self.inventory.open()
        self.inventory.drop_all(items)
        self.inventory.close()

    def spam_hotbar(self):
        """Typewrites all the hotbar keys with crystals on them to open crystals fast."""
        pg.typewrite("".join(c for c in self.HOTBAR), interval=0.01)

    def set_hotbar(self) -> None:
        """Sets the hotbar using shift left click on the crystals"""
        input.keyDown("shift")
        for _ in range(4):
            for slot in self.HOTBAR:
                self.press(slot)
                self.sleep(0.5)
        input.keyUp("shift")

    def walk(self, key, duration):
        """'Walks' the given direction for the given duration.

        Parameters:
        ----------
        key :class:`str`:
            The key to hold to walk

        duration :class: `int`|`float`:
            The duration to walk for
        """
        input.keyDown(key)
        self.sleep(duration=duration)
        input.keyUp(key)

    def crouch(self) -> None:
        """Crouches the player"""
        self.press(self.keybinds.crouch)

    def prone(self) -> None:
        """Prones the player"""
        self.press(self.keybinds.prone)

    def stand_up(self) -> None:
        for _ in range(3):
            input.press("shift")
        input.keyUp("shift")
    def disable_hud(self) -> None:
        """Disables HUD"""
        self.press("backspace")

    def pick_up_bag(self):
        """Picks up items from a drop script bag, deletes the bag after."""
        self.look_down_hard()
        self.press(self.keybinds.target_inventory)
        if await_event(self.received_item, max_duration=3):
            self.sleep(0.5)
        self._popcorn_bag()

    def set_first_person(self) -> None:
        """Sets the player to first person"""
        self.mouse_scroll(1)

    def hide_hands(self) -> None:
        """Looks up and down to hide the hands temporarily"""
        self.look_up_hard()
        self.sleep(0.2)
        self.look_down_hard()
        self.disable_hud()
        self.sleep(0.5)
        self.turn_y_by(-160)
        self.sleep(0.3)

    def received_item(self) -> bool:
        """Returns whether the player has received an item. Not to be confused
        with the `received_item` method of the `Inventory` class, which does
        the same but on a different position."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/added.png",
                region=self._ADDED_REGION,
                confidence=0.75,
            )
            is not None
        )

    def has_died(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/you_died.png",
                region=self._HAS_DIED,
                confidence=0.7,
            )
            is not None
        )

    def is_spawned(self) -> bool:
        """Checks if the player is spawned"""
        return self.hud.is_open()

    def has_buff(self, buff: Buff) -> bool:
        """Checks if the player has the given buff"""
        return (
            self.window.locate_template(
                buff.image, region=self._DEBUFF_REGION, confidence=0.8
            )
            is not None
        )

    def needs_recovery(self) -> bool:
        """Checks if the player needs to recover"""
        return any(self.has_buff(buff) for buff in [THIRSTY, HUNGRY, BROKEN_BONES])

    def do_drop_script(self, item: Item, target: Inventory):
        """Does the item drop script for the given item in the given structure.
        Used to empty heavy items out of structures that are not dedis. Player
        has to be non crouching and will end up as not crouching.

        Parameters:
        -----------
        item :class:`Item`:
            The item to dropscript for

        target :class:`Inventory`:
            The inventory to take the item out of.

        slot :class: `int`:
            The slot to take the item from, required if the inventory has
            an item folder in slot 1.
        """
        self.crouch()
        self.sleep(0.5)

        target.open(max_duration=45)

        target.take(item, amount=1)
        self.inventory.await_items_added(item)
        self.sleep(0.3)

        self.inventory.drop_all([item])
        self.inventory.search(item)

        while target.has(item, is_searched=True):
            target.transfer_all()
            target.search(item)
            self.sleep(0.2)

            if not target.has(item, is_searched=True):
                break

            self.inventory.drop(item, search=False)
            target.search(item)

        self.inventory.close()
        self.pick_up_bag()
        self.sleep(1)
        self.stand_up()

    def spawn_in(self) -> None:
        """Waits for the player to spawn in, catching possible scenarios such
        as the player dying instantly upon spawning.

        Parameters
        ---------
        screen :class:`SpawnScreen`:
            The spawn screen used to travel
        """
        start = time.time()
        while not self.is_spawned():
            pg.keyUp(self.keybinds.hud_info)
            pg.keyDown(self.keybinds.hud_info)
            self.sleep(0.1)

            if timedout(start, 15):
                pg.keyUp(self.keybinds.hud_info)
                return

            if self.has_died():
                raise PlayerDiedError("Spawning")

        pg.keyUp(self.keybinds.hud_info)

    def _is_travelling(self) -> bool:
        """Check if we are currently travelling (whitescreen)"""
        return pg.pixelMatchesColor(
            *self.window.convert_point(959, 493), (255, 255, 255), tolerance=10
        )

    def _popcorn_bag(self) -> None:
        bag = Structure("Item Cache", "assets/wheels/item_cache.png")
        bag.open()
        bag.inventory.select_slot(0)

        while bag.inventory.is_open():
            self.press("o")
            self.sleep(0.3)



from datetime import datetime
from typing import Literal, Optional, final

from ..._helpers import format_seconds
from .structure import Structure


@final
class IndustrialForge(Structure):
    """Represents the Industrial Forge in Ark.

    The industrial forge extends the abilities the `Structure` class, by adding
    methods to keep track of different types of items burning and checking on 
    how much longer is left on them to finish.

    Beware that, while you may consider charcoal, metal ingots or gasoline
    to be craftables of the `Industrial Forge`, they are not considered as such.
    It is defined as a toggleable, but items that are cooked in the forge are
    not technically crafted.

    Parameters
    ----------
    burning_item :class:`Literal["wood", "metal", "oil"]`:
        The item being cooked in the forge.

    started_cooking :class:`Optional[datetime]`:
        The time the forge started cooking, if not passed it will be assumed the
        forge never cooked.

    Properties
    ----------
    burning_item :class:`str`:
        The item being burnt, can be changed at runtime but with the same
        limitations as during the initialization.

    is_done_burning :class:`bool`:
        Whether the forge is done burning, taking into account the burning time
        of the item it is burning.
    """

    BURNING_DURATIONS: dict[str, float | int] = {"wood": 2.8, "metal": 4.3, "oil": 0.7}

    def __init__(
        self,
        burning_item: Literal["wood", "metal", "oil"] = "wood",
        started_cooking: Optional[datetime] = None,
    ) -> None:
        if burning_item not in ["wood", "metal", "oil"]:
            raise ValueError(
                f'Expected one of {["wood", "metal", "oil"]}, got "{burning_item}".'
            )
        super().__init__(
            name="Industrial Forge",
            action_wheel="assets/wheels/industrial_forge.png",
            capacity="assets/interfaces/forge_full.png",
            toggleable=True,
        )
        self.started_cooking = started_cooking
        self._burning_item = burning_item

    @property
    def burning_item(self) -> str:
        return self._burning_item

    @burning_item.setter
    def burning_item(self, item: Literal["wood", "metal", "oil"]) -> None:
        if not item in ["wood", "metal", "oil"]:
            raise ValueError(
                f'Cant burn {item}, expected one of {["wood", "metal", "oil"]}'
            )
        self._burning_item = item

    @property
    def is_done_burning(self) -> bool:
        if self.started_cooking is None:
            return True
        delta = datetime.now() - self.started_cooking
        return delta.total_seconds() > self.BURNING_DURATIONS[self.burning_item] * 3600

    @property
    def remaining_burn_time(self) -> str:
        if self.started_cooking is None:
            return "0"

        delta = datetime.now() - self.started_cooking
        duration = self.BURNING_DURATIONS[self.burning_item]
        
        remaining = (duration * 3600) - delta.total_seconds()
        return format_seconds(round(remaining))


from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Optional

import dacite

from .. import config


@dataclass
class DinoExport:

    dino_name: str
    tamed_name: str
    health: float
    stamina: float
    oxygen: float
    food: float
    weight: float
    melee: float
    speed: float
    crafting: float

    @staticmethod
    def load_most_recent(path: Optional[str] = None) -> DinoExport:
        """Loads the settings from input.ini, using the `ARK_PATH` provided
        in the configs or an alternatively passed path."""
        if path is None:
            path = f"{config.ARK_PATH}\ShooterGame\Saved\DinoExports"
            
        if "common" in path:
            id_folder = max(
                [p for p in Path(path).iterdir()],
                key=lambda p: datetime.fromtimestamp(p.stat().st_mtime).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
            )
            path += f"\{id_folder.stem}"

        file = max(
            [p for p in Path(path).iterdir()],
            key=lambda p: datetime.fromtimestamp(p.stat().st_mtime).strftime(
                "%Y-%m-%d %H:%M:%S"
            ),
        )
        settings: dict[str, str | float] = {}
        try:
            with open(file, encoding="utf-8") as f:
                contents = f.readlines()
        except UnicodeDecodeError:
            with open(file, encoding="utf-16") as f:
                contents = f.readlines()

        for line in contents:
            line = line.strip()
            if line.count("=") != 1:
                continue

            key, val = line.split("=")
            action = _KEY_MAP.get(key)
            if action is None:
                continue
            try:
                settings[action] = float(val)
            except ValueError:
                settings[action] = val

        return dacite.from_dict(DinoExport, settings)


_KEY_MAP = {
    "DinoNameTag": "dino_name",
    "TamedName": "tamed_name",
    "Health": "health",
    "Stamina": "stamina",
    "Oxygen": "oxygen",
    "food": "food",
    "Weight": "weight",
    "Melee Damage": "melee",
    "Movement Speed": "speed",
    "Crafting Skill": "crafting",
}



from typing import Literal, final

from .._button import Button
from .inventory import Inventory


@final
class DedicatedStorageInventory(Inventory):
    """Represents the Dedicated Storage in Ark.

    Extends the abilities of a regular `Inventory` by providing control
    over the additional buttons a dedicated storage adds to the interface.
    """

    _DEPOSIT_ALL = Button((962, 618))
    _WITHDRAW_STACK = Button((962, 660))
    _CLEAR_RESOURCE = Button((962, 703))

    _WITHDRAW_ONE = Button((962, 770))
    _WITHDRAW_FIVE = Button((962, 822))
    _WITHDRAW_TEN = Button((962, 877))

    def __init__(self):
        super().__init__("Tek Dedicated Storage", "assets/wheels/dedi.png")

    def deposit(self) -> None:
        self.click_at(self._DEPOSIT_ALL.location)

    def withdraw_stacks(self, amount: int) -> None:
        for _ in range(amount):
            self.click_at(self._WITHDRAW_STACK.location)

    def withdraw(self, amount: Literal[1, 5, 10], presses: int = 1) -> None:
        buttons = {
            1: self._WITHDRAW_ONE,
            5: self._WITHDRAW_FIVE,
            10: self._WITHDRAW_TEN,
        }
        for _ in range(presses):
            self.click_at(buttons[amount].location)



import pyautogui as pg  # type: ignore[import]

from .._ark import Ark
from .._helpers import await_event


class Console(Ark):
    """Represents the Console in Ark.
    
    Provides the ability to enter commands such as t.maxfps, gamma...
    """

    COMMANDS = ["t.maxfps", "gamma", "disconnect", "reconnect", "exit"]

    def is_open(self) -> bool:
        """Returns whether the console is open by matching the black par"""
        return pg.pixelMatchesColor(
            *self.window.convert_point(976, 1071), (0, 0, 0), tolerance=3
        )

    def open(self):
        """Opens the console, times out after 10 seconds"""
        attempts = 0
        while not self.is_open():
            self.press(self.keybinds.console)
            if await_event(self.is_open, max_duration=3):
                return
            attempts += 1
            if attempts > 3:
                raise TimeoutError("Failed to open the console")

    def run(self, command: str) -> None:
        """Executes a given command in the console."""
        self.open()

        pg.typewrite(command, interval=0.001)
        self.sleep(0.5)
        self.press("enter") 

    def set_fps(self, fps: int | str = 10):
        """Sets the fps to a given value"""
        self.open()

        pg.typewrite(f"t.maxfps {fps}", interval=0.001)
        self.sleep(0.5)
        self.press("enter")

    def set_gamma(self, gamma: int | str = 5):
        """Sets gamma to a given value"""
        self.open()

        pg.typewrite(f"gamma {gamma}", interval=0.001)
        self.sleep(0.5)
        self.press("enter")

    def run_required_commands(self):
        """Sets up max fps and gamma with default parameters"""
        self.set_fps()
        self.set_gamma()



import cv2 as cv  # type:ignore[import]
import numpy as np
import pyautogui  # type:ignore[import]

from ..._helpers import find_center, find_closest_pixel, get_center
from ...exceptions import ActionNotFoundError
from .wheel import ActionWheel


class StryderWheel(ActionWheel):
    def __init__(self) -> None:
        super().__init__("Stryder", "assets/wheels/stryder.png")

    def enter_dedi_transfer_tab(self) -> None:
        self.sleep(1)
        position = self.find_dedi_transfer_tab()
        if position is None:
            raise ActionNotFoundError("Dedi transfer")

        self.sleep(0.3)
        self.select_action(position)
        self.sleep(1)

    def enter_linked_dedi_tab(self) -> None:
        self.select_action((1145, 783))
        self.sleep(0.5)

    def transfer_to_nearby_dedis(self) -> None:
        self.select_action((960, 230))
        pyautogui.keyUp("e")

    def show_more(self) -> None:
        self.select_action((835, 263))

    def find_dedi_transfer_tab(self) -> tuple[int, int] | None:
        """Finds the dedi transfer tab by denoising for the orange text.

        Matching for the orange text template is not reliable because the text
        size changes depending on how many options are on the action wheel.

        Returns the coordinate of the transfer tab as a tuple, or None if not found.
        """
        img_arr = np.array(self.window.grab_screen(self._WHEEL_AREA))
        img = cv.cvtColor(img_arr, cv.COLOR_BGR2RGB)

        lower_bound = tuple(max(0, i - 15) for i in (255, 146, 39))
        upper_bound = tuple(min(255, i + 15) for i in (255, 146, 39))

        mask = cv.inRange(img, lower_bound, upper_bound)
        matches = cv.findNonZero(mask)
        if matches is None:
            return None

        matching_points = [
            (self._WHEEL_AREA[0] + x, self._WHEEL_AREA[1] + y)
            for x, y in np.concatenate(matches, axis=0)
        ]
        com = find_center(matching_points)
        return find_closest_pixel(matching_points, com)


from typing import final

from ..spawn_screen import SpawnScreen
from .structure import Structure


@final
class Bed(Structure):

    def __init__(self, name: str) -> None:
        super().__init__(name, "assets/wheels/bed.png")
        self.interface = SpawnScreen()

    def spawn(self) -> None:
        self.interface.travel_to(self.name)


    def lay_down(self) -> None:
        self.action_wheel.activate()
        self.action_wheel.select_action((1130, 510), click=False)

    def get_up(self) -> None:
        self.press(self.keybinds.use)



from typing import final

from ..inventories import CropPlotInventory
from .structure import Structure


@final
class TekCropPlot(Structure):
    
    def __init__(self, name: str) -> None:
        super().__init__(name, "assets/wheels/tek_crop_plot.png", inventory=CropPlotInventory(name))


from .._ark import Ark


class MainMenu(Ark):
    """Represents the Main Menu in Ark.

    Provides the ability to recognize whether the player has disconnected.
    """

    _OPTIONS = (20, 600, 230, 76)
    _ACCEPT = (515, 320, 910, 390)

    def is_open(self) -> bool:
        """Returns whether the main menu is currently open."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/main_menu_options.png",
                region=self._OPTIONS,
                confidence=0.8,
            )
            is not None
        )

    def player_disconnected(self) -> bool:
        """Returns whether the player was disconnected."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/main_menu_accept.png",
                region=self._ACCEPT,
                confidence=0.8,
            )
            is not None
        )



import cv2  # type:ignore[import]
import pyautogui  # type: ignore[import]
import pydirectinput  # type: ignore[import]
from pytesseract import pytesseract as tes  # type: ignore[import]

from ..._ark import Ark
from ..._helpers import (await_event, find_center, find_closest_pixel,
                       get_center, get_filepath)
from ...exceptions import UnexpectedWheelError, WheelNotAccessibleError


class ActionWheel(Ark):
    """Represents the action wheel in Ark.

    Provides the ability to select certain tabs on the action wheel,
    check whether it is open, get the wheels text or find certain tabs.

    Wheels are useful because they help figure out if we just simply
    arent looking at a structure while trying to access it, or if lag
    is holding us back from opening it, because action wheels open regardless
    of lag or timer.

    Parameters
    ----------
    filepath :class:`str`:
        The filepath to an image of the action wheel
    """

    _WHEEL_NAME_AREA = (840, 425, 240, 230)
    _WHEEL_AREA = (543, 135, 867, 825)

    def __init__(self, name: str, filepath: str) -> None:
        super().__init__()
        self._name = name
        self._filepath = get_filepath(filepath)

    def __str__(self) -> str:
        return f"Action Wheel '{self._name}' of type '{type(self).__name__}' with filepath {self._filepath}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={self._name}, filepath={self._filepath})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def filepath(self) -> str:
        return self._filepath

    def activate(self) -> None:
        """Brings up the action wheel. When the function ends, the 'E' key
        is still being held for further usage of the wheel. Remember to release
        it yourself when you need it gone.

        Raises
        ------
        `UnexpectedWheelError`
            If after 3 attempts a wheel is open but not the expected one

        `WheelNotAccessibleError`
            If no wheel could be opened after 5 attempts
        """
        attempts = 0
        while not self.is_open():
            pyautogui.keyDown(self.keybinds.use)
            if await_event(self.is_open, max_duration=2):
                break

            attempts += 1
            if attempts > 3:
                raise WheelNotAccessibleError(self._name)

            other_text = self.get_text()
            if len(other_text) > 5:
                raise UnexpectedWheelError(self._name, other_text)
                
            self.sleep(1)
            pyautogui.keyUp(self.keybinds.use)

    def deactivate(self) -> None:
        pyautogui.keyUp(self.keybinds.use)
        self.sleep(0.5)
        
    def select_action(self, position: tuple[int, int], click: bool = True) -> None:
        pydirectinput.moveTo(*position, duration=0.1)
        self.sleep(0.1)
        pydirectinput.moveTo(*position, duration=0.1)
        self.sleep(1)
        if click:
            self.click("left")

    def find_action(self, whitelist: str) -> tuple[int, int] | None:
        wheel = self.window.grab_screen(self._WHEEL_AREA)
        wheel_outer: cv2.Mat = self.window.denoise_text(
            wheel, (255, 255, 255), variance=14, dilate=False
        )

        height, _ = wheel_outer.shape

        boxes: str = tes.image_to_boxes(
            wheel_outer, config=f"-c tessedit_char_whitelist={whitelist} --psm 3"
        )
        points = []
        for box in boxes.splitlines():
            _, x, y, w, h, _ = box.split(" ")
            center = get_center(
                (
                    int(x),
                    height - int(y),
                    int(w) - int(x),
                    (height - int(h)) - (height - int(y)),
                )
            )
            points.append(
                (center[0] + self._WHEEL_AREA[0], center[1] + self._WHEEL_AREA[1])
            )

        if not points:
            return None

        com = find_center(points)
        return find_closest_pixel(points, com)

    def get_text(self) -> str:
        """Returns the text of the action wheel."""
        wheel_inner = self.window.grab_screen(self._WHEEL_NAME_AREA)
        wheel_inner_text = self.window.denoise_text(
            wheel_inner, (255, 209, 64), variance=7
        )
        raw: str = tes.image_to_string(
            wheel_inner_text,
            config="--psm 6 -l eng",
        )
        return raw.replace("\n", " ").replace("{", "(").replace("}", "}")

    def is_open(self) -> bool:
        """Returns whether the action wheel is currently open"""
        return (
            self.window.locate_template(
                self._filepath,
                region=self._WHEEL_NAME_AREA,
                confidence=0.7,
            )
            is not None
        )

    def in_access_range(self) -> bool:
        """Uses the action wheel to check if a structure or dinosaur is within
        access range to determine if we are unable to open it because of lag or
        if we are just not looking at it. Useful for time efficiency on crop plots.

        Returns `True` if the action wheel displays the structure / dinos name.
        """
        with pyautogui.hold("e"):
            self.sleep(1)
            return self.is_open()

    def export_data(self) -> None:
        self.activate()
        self.select_action((811, 470), click=True)
        self.sleep(1)
        self.select_action((850, 642), click=True)

        self.deactivate()


from .buff import Buff

HUNGRY = Buff("Hungry", "assets/buffs/hungry.png")
THIRSTY = Buff("Thirsty", "assets/buffs/thirsty.png")
POD_XP = Buff("XP Buff", "assets/buffs/pod_buff.png")
BROKEN_BONES = Buff("Broken Bones", "assets/buffs/broken_bones.png")


from typing import final

from ...items import GUNPOWDER, SPARKPOWDER
from .structure import Structure


@final
class ChemistryBench(Structure):
    """Represents the chem bench in ark.
    Is able to be turned on and off.
    """

    def __init__(self) -> None:
        super().__init__(
            "Chemistry Bench",
            "assets/wheels/chemistry_bench.png",
            craftables=[SPARKPOWDER, GUNPOWDER],
            toggleable=True
        )







from .crop_plot_inventory import CropPlotInventory
from .dedi_inventory import DedicatedStorageInventory
from .inventory import Inventory
from .player_inventory import PlayerInventory

__all__ = ("Inventory", "DedicatedStorageInventory", "PlayerInventory", "CropPlotInventory")



from ctypes.wintypes import CHAR
from .item import Item

DUST = Item(
    name="Element Dust",
    search_name="dust",
    stack_size=1000,
    inventory_icon="assets/items/resources/dust.png",
    added_icon="assets/items/dust_deposited.png",
    added_text="assets/items/dust_text.png",
)

BLACK_PEARL = Item(
    name="Black Pearl",
    search_name="black",
    stack_size=200,
    inventory_icon="assets/items/resources/black_pearl.png",
    added_icon="assets/items/black_pearl_deposited.png",
    added_text="assets/items/black_pearl_text.png",
)

CRYSTAL = Item(
    name="Crystal",
    search_name="crystal",
    stack_size=100,
    inventory_icon="assets/items/resources/crystal.png",
)

GREENGEM = Item(
    name="Fragmented Green Gem",
    search_name="green gem",
    stack_size=100,
    inventory_icon="assets/items/resources/green_gem.png",
)

SAND = Item(
    name="Sand",
    search_name="sand",
    stack_size=100,
    inventory_icon="assets/items/resources/sand.png",
)

THATCH = Item(
    name="Thatch",
    search_name="thatch",
    stack_size=200,
    inventory_icon="assets/items/resources/thatch.png",
)

BLUEGEM = Item(
    name="Blue Crystalized Sap",
    search_name="blue",
    stack_size=100,
    inventory_icon="assets/items/resources/blue_gem.png",
)

SILK = Item(
    name="Silk",
    search_name="silk",
    stack_size=200,
    inventory_icon="assets/items/resources/silk.png",
)

CLAY = Item(
    name="Clay",
    search_name="clay",
    stack_size=100,
    inventory_icon="assets/items/resources/clay.png",
)

OIL = Item(
    name="Oil",
    search_name="oil",
    stack_size=100,
    inventory_icon="assets/items/resources/oil.png",
)

GASBALL = Item(
    name="Congealed Gas Balls",
    search_name="gas ball",
    stack_size=100,
    inventory_icon="assets/items/resources/gasball.png",
)

SULFUR = Item(
    name="Sulfur",
    search_name="sulfur",
    stack_size=100,
    inventory_icon="assets/items/resources/sulfur.png",
)

METAL = Item(
    name="Metal",
    search_name="metal",
    stack_size=200,
    inventory_icon="assets/items/resources/metal.png",
)

REDGEM = Item(
    name="Red Crystalized Sap",
    search_name="red",
    stack_size=100,
    inventory_icon="assets/items/resources/red_gem.png",
)

OBSIDIAN = Item(
    name="Obsidian",
    search_name="obsidian",
    stack_size=100,
    inventory_icon="assets/items/resources/obsidian.png",
)

SAP = Item(
    name="Sap",
    search_name="sap",
    stack_size=100,
    inventory_icon="assets/items/resources/sap.png",
)

METAL_INGOT = Item(
    name="Ingot",
    search_name="ingot",
    stack_size=300,
    inventory_icon="assets/items/resources/metal_ingot.png",
    added_icon="assets/items/ingot_icon.png",
    added_text="assets/items/ingot_text.png",
)

FLINT = Item(
    name="Flint",
    search_name="flint",
    stack_size=100,
    inventory_icon="assets/items/resources/flint.png",
)

STONE = Item(
    name="Stone",
    search_name="stone",
    stack_size=100,
    inventory_icon="assets/items/resources/stone.png",
    added_icon="assets/items/stone_icon.png",
    added_text="assets/items/stone_text.png",
)

FUNGAL_WOOD = Item(
    name="Fungal wood",
    search_name="wood",
    stack_size=100,
    inventory_icon="assets/items/resources/fungal_wood.png",
)

PASTE = Item(
    name="Paste",
    search_name="paste",
    stack_size=100,
    inventory_icon="assets/items/resources/paste.png",
    added_icon="assets/items/paste_icon.png",
    added_text="assets/items/paste_text.png",
)

SILICA_PEARL = Item(
    name="Silica Pearl",
    search_name="pearls",
    stack_size=100,
    inventory_icon="assets/items/resources/silica_pearl.png",
    added_icon="assets/items/pearls_icon.png",
    added_text="assets/items/pearls_text.png",
)

HIDE = Item(
    name="Hide",
    search_name="hide",
    stack_size=200,
    inventory_icon="assets/items/resources/hide.png",
)

HARD_POLYMER = Item(
    name="Polymer",
    search_name="poly",
    stack_size=100,
    inventory_icon="assets/items/resources/polymer.png",
)

ORGANIC_POLYMER = Item(
    name="Organic Polymer",
    search_name="poly",
    stack_size=20,
    inventory_icon="assets/items/resources/organic_polymer.png",
)

GASOLINE = Item(
    name="Gasoline",
    search_name="gasoline",
    stack_size=100,
    inventory_icon="assets/items/resources/gasoline.png",
)

CHARCOAL = Item(
    name="Charcoal",
    search_name="coal",
    stack_size=100,
    inventory_icon="assets/items/resources/charcoal.png",
)

SPARKPOWDER = Item(
    name="Sparkpowder",
    search_name="spark",
    recipe={FLINT: 2, STONE: 1},
    stack_size=100,
    inventory_icon="assets/items/resources/sparkpowder.png",
)

GUNPOWDER = Item(
    name="Gunpowder",
    search_name="gunpowder",
    recipe={SPARKPOWDER: 1, CHARCOAL: 1},
    stack_size=100,
    inventory_icon="assets/items/resources/gunpowder.png",
)

ARB = Item(
    name="Advanced Rifle Bullet",
    search_name="advanced rifle bullet",
    stack_size=100,
    inventory_icon="assets/items/resources/arb.png",
    added_icon="assets/items/arb_icon.png",
    added_text="assets/items/arb_text.png",
)

ANGLER_GEL = Item(
    name="Angler Gel",
    search_name="gel",
    stack_size=100,
    inventory_icon="assets/items/resources/angler_gel.png",
)

FIBER = Item(
    name="Fiber",
    search_name="fiber",
    stack_size=300,
    inventory_icon="assets/items/resources/fiber.png",
)

WOOD = Item(
    name="Wood",
    search_name="wood",
    stack_size=100,
    inventory_icon="assets/items/resources/wood.png",
)

Y_TRAP = Item(
    name="Y Trap",
    search_name="trap",
    stack_size=10,
    inventory_icon="assets/items/resources/ytrap.png",
    added_text="assets/items/y_trap_added.png",
    added_icon="assets/items/plant_species_y_trap.png",
)

GACHA_CRYSTAL = Item(
    name="Gacha Crystal",
    search_name="gacha",
    stack_size=1,
    inventory_icon="assets/items/resources/gacha_crystal.png",
)

ELECTRONICS = Item(
    name="Electronics",
    search_name="electronics",
    stack_size=100,
    recipe={SILICA_PEARL: 3, METAL_INGOT: 1},
    inventory_icon="assets/items/resources/electronics.png",
    added_icon="assets/items/electronics_icon.png",
    added_text="assets/items/electronics_text.png",
)

ELEMENT = Item(
    name="Element",
    search_name="element",
    stack_size=100,
    inventory_icon="assets/items/resources/element.png",
)


import cv2 as cv  # type:ignore[import]
import numpy as np
import pyautogui  # type:ignore[import]

from .._ark import Ark
from .._helpers import await_event, get_center
from ..exceptions import BedNotAccessibleError, BedNotFoundError, PlayerDidntTravelError

from .. import config


class SpawnScreen(Ark):
    """Represents the spawn screen in Ark.

    Provides the ability to search for beds and detect them on the map,
    to then travel to them. If the bed icon itself cannot be found, the
    center of the red X will be assumed to be the correct location.
    """

    SEARCH_BAR = (306, 982)
    SPAWN_BUTTON = (731, 978)

    _BEDS_REGION = (160, 70, 1050, 880)
    _BED_NAME_AREA = (624, 967, 250, 25)

    def spawn(self) -> None:
        """Clicks the spawn button"""
        self.click_at(self.SPAWN_BUTTON)

    def search(self, name: str) -> None:
        """Searches for a bed"""
        self.click_at(self.SEARCH_BAR)
        with pyautogui.hold("ctrl"):
            pyautogui.press("a")

        pyautogui.typewrite(name.lower(), interval=0.001)
        self.sleep(0.3)

    def open(self) -> None:
        """Opens the bed menu. Times out after 30 unsuccessful
        attempts raising a `BedNotAccessibleError`"""
        attempt = 0
        while not self.is_open():
            attempt += 1
            self.press(self.keybinds.use)
            self.sleep(1)

            if attempt > 30:
                raise BedNotAccessibleError("Failed to access the bed!")

    def travel_to(self, bed_name: str) -> None:
        """Travels to a bed given it's name. If the spawn screen is not
        already open, it will be opened first.

        Parameters
        ----------
        name :class:`str`:
            The name of the bed to travel to
        """
        self.open()
        self.search(bed_name)

        position = self._find_bed() or self._find_x()
        if position is None:
            raise BedNotFoundError(f"Could not find '{bed_name}'!")

        self.click_at(position, delay=0.5)
        self.sleep(0.5)

        attempts = 0
        while not self._bed_is_selected():
            self.click_at(position, delay=0.5)
            if await_event(self._bed_is_selected, max_duration=1):
                break

            attempts += 1
            if attempts >= 2:
                raise BedNotFoundError(f"Could not select '{bed_name}'!")

            self.search(bed_name)
            position = self._find_bed() or self._find_x()

        if self._spawn_region_is_selected():
            raise BedNotFoundError(
                f"Could not find '{bed_name}'! Random spawn region was selected!"
            )

        self.spawn()
        if await_event(self._is_travelling, max_duration=15 * config.TIMER_FACTOR):
            self.sleep(2)
            return
        raise PlayerDidntTravelError(f"Failed to travel to bed '{bed_name}'!")

    def can_be_accessed(self) -> bool:
        """Returns whether the bed can be accessed, determined by the
        'Fast trave' text that appears when facing it."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/fast_travel.png",
                region=(0, 0, 1920, 1080),
                confidence=0.7,
            )
            is not None
        )

    def is_open(self) -> bool:
        """Returns whether the spawn screen is currently open."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces//bed_filter.png",
                region=(140, 950, 150, 50),
                confidence=0.8,
            )
            is not None
        )

    def _find_bed(self) -> tuple[int, int] | None:
        """Finds the icon of a bed on the map, assuming the bed has already
        been searched."""
        img_arr = np.array(self.window.grab_screen(self._BEDS_REGION))
        img = cv.cvtColor(img_arr, cv.COLOR_BGR2RGB)

        lower_bound = tuple(max(0, i - 7) for i in (0, 255, 255))
        upper_bound = tuple(min(255, i + 7) for i in (0, 255, 255))

        mask = cv.inRange(img, lower_bound, upper_bound)
        contours, _ = cv.findContours(
            mask.copy(), cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE
        )

        if len(contours) == 0:
            return None

        contour = max(contours, key=cv.contourArea)
        if cv.contourArea(contour) < 10:
            return None

        point = get_center(cv.boundingRect(contour))
        return point[0] + 160, point[1] + 70

    def _find_x(self) -> tuple[int, int] | None:
        """Finds the red X on the map, assuming the bed has already
        been searched."""
        img_arr = np.array(self.window.grab_screen(self._BEDS_REGION))
        img = cv.cvtColor(img_arr, cv.COLOR_BGR2RGB)

        lower_bound = tuple(max(0, i - 7) for i in (255, 255, 255))
        upper_bound = tuple(min(255, i + 7) for i in (255, 255, 255))

        mask = cv.inRange(img, lower_bound, upper_bound)
        contours, _ = cv.findContours(
            mask.copy(), cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE
        )

        if len(contours) == 0:
            return None

        contour = max(contours, key=cv.contourArea)
        if cv.contourArea(contour) < 300:
            return None

        point = get_center(cv.boundingRect(contour))
        return point[0] + 160, point[1] + 70

    def _is_travelling(self) -> bool:
        """Check if we are currently travelling (whitescreen)"""
        return pyautogui.pixelMatchesColor(
            *self.window.convert_point(959, 493), (255, 255, 255), tolerance=10
        )

    def _bed_is_selected(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/bed_button.png",
                region=(587, 955, 35, 43),
                confidence=0.7,
            )
            is not None
        )

    def _spawn_region_is_selected(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/random_location.png",
                region=self._BED_NAME_AREA,
                confidence=0.85,
            )
            is not None
        )



INVENTORY_OPEN_INTERVAL: int | float = 5
INVENTORY_CLOSE_INTERVAL: int | float = 5
TIMER_FACTOR: int | float = 1
ARK_PATH: str = "F:\ARKSurvivalEvolved"
TESSERACT_PATH: str = r"C:\Program Files\Tesseract-OCR\tesseract.exe"


import time
from typing import final

from ... import config
from ..._helpers import timedout
from ...entities.player import Player
from ..spawn_screen import SpawnScreen
from ..wheels import TekPodWheel
from .structure import Structure


@final
class TekSleepingPod(Structure):
    """Represents the Tek Sleeping Pod in Ark.
    
    Provides ability to enter the pod to heal the player and access
    to the `TekPodWheel` class.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name, TekPodWheel())
        self.action_wheel: TekPodWheel
        self.interface = SpawnScreen()
        
    def spawn(self) -> None:
        self.interface.travel_to(self.name)

    def heal(self, player: Player) -> None:
        """Heals the player until either all relevant stats are full,
        or the calculated duration has expired."""
        player.stand_up()
        player.look_down_hard()

        self.action_wheel.activate()
        self.action_wheel.lay_on()

        player.inventory.open()
        self._wait_to_heal(player)
        player.inventory.close()

    def _wait_to_heal(self, player: Player) -> None:
        """Waits to heal the player given its stats or the state
        of the stats."""
        hp_duration = player.stats.health / 10
        water_duration = player.stats.water / 2.5
        food_duration = player.stats.food / 2.5

        max_duration = max(hp_duration, water_duration, food_duration) * config.TIMER_FACTOR
        start = time.time()
        while not all(
            (
                player.inventory.hp_full(),
                player.inventory.water_full(),
                player.inventory.food_full(),
            )
        ):
            self.sleep(1)
            if timedout(start, max_duration):
                break

    def leave(self) -> None:
        """Leaves the tek pod"""
        self.press(self.keybinds.use)


import time

from .._ark import Ark
from .._helpers import await_event, timedout
from ..exceptions import InterfaceError


class EscapeMenu(Ark):
    """Represents the Escape Menu in Ark.

    Provides the ability to open or close it, which is useful for clearing the screen
    i.e close unwanted interfaces.
    """

    _RESUME_REGION = (750, 215, 425, 200)

    def open(self) -> None:
        """Opens the escape menu."""
        start = time.time()

        while not self.is_open():
            self.press("escape")
            if await_event(self.is_open, max_duration=3):
                return

            if timedout(start, 30):
                raise InterfaceError("Failed to open escape menu!")

    def close(self) -> None:
        """Closes the escape menu."""
        start = time.time()

        while self.is_open():
            self.press("escape")
            if await_event(self.is_open, False, max_duration=3):
                return

            if timedout(start, 30):
                raise InterfaceError("Failed to close escape menu!")

    def is_open(self) -> bool:
        """Checks if the menu is open."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/resume.png",
                region=self._RESUME_REGION,
                confidence=0.8,
            )
            is not None
        )



import math
import time
from typing import Iterable, Literal, Optional, final, overload

import pyautogui as pg  # type: ignore[import]
from pytesseract import pytesseract as tes  # type: ignore[import]

from ... import config
from ..._ark import Ark
from ..._helpers import (await_event, get_center, get_filepath, set_clipboard,
                         timedout)
from ...exceptions import (InventoryNotAccessibleError,
                           InventoryNotClosableError, InventoryNotOpenError,
                           NoItemsAddedError, ReceivingRemoveInventoryTimeout,
                           UnknownFolderIndexError)
from ...items import Item
from .._button import Button


class Inventory(Ark):
    """Represents an inventory in Ark.

    All `Structure` and `Dinosaur` objects contain an `Inventory` attribute,
    it is responsible for all action related to transferring items, clicking
    buttons within the inventory and using the searchbar.

    For inventories that contain `craftables`, it also provides the methods
    needed to use the crafting tab and craft requested items. Furthermore,
    it is able to keep track and sync it's contents.

    Parameters:
    -----------
    entity :class:`str`:
        The name of the structure / dino the inventory belongs to

    craftables :class:`list[Item]` [Optional]:
        A list of items that can be crafted in the in the crafting tab

    capacity :class:`str`: [Optional]
        An image path containing the image of the max capacity.

    Properties:
    ----------
    contents :class:`dict[Item, int]`:
        A dictionary mapping items in the structure to their quantity

    capacity :class:`int`:
        The set maximum slots the inventory can hold

    craftables :class:`list[Item]`:
        A list of items that can be crafted in the inventories crafting tab.
    """

    LAST_TRANSFER_ALL = time.time()
    SLOTS = [
        (x, y, 93, 93) for y in range(232, 883, 93) for x in range(1243, 1708 + 93, 93)
    ]
    LEVEL_UP_BUTTONS = {
        "health": (1150, 515),
        "stamina": (1150, 558),
        "oxygen": (1150, 600),
        "food": (1151, 645),
        "weight": (1149, 688),
        "melee": (1150, 730),
        "speed": (1151, 773),
        "crafting": (1149, 818),
    }
    CRAFTING_QUEUE = (1234, 779, 571, 168)

    _FOLDERS = [
        "AAA",
        "BBB",
        "CCC",
        "DDD",
        "EEE",
        "FFF",
        "GGG",
        "HHH",
        "III",
        "JJJ",
        "KKK",
    ]

    _FOLDER_VIEW = Button((1663, 188), (1632, 158, 61, 56), "folder_view.png")
    _SHOW_ENGRAMS = Button((1716, 189), (1690, 160, 51, 51), "show_engrams.png")
    _UNL_ENGRAMS = Button((1770, 188), (1742, 160, 56, 54), "unlearned_engrams.png")
    _TRANSFER_ALL = Button((1425, 190))
    _DROP_ALL = Button((1477, 187))
    _CRAFTING_TAB = Button((1716, 118), (1627, 82, 211, 69), "crafting.png")
    _INVENTORY_TAB = Button((1322, 118), (1235, 88, 184, 60), "inventory.png")
    _CREATE_FOLDER = Button((1584, 187))
    _STOP_CRAFTING = Button((1775, 800), (1750, 775, 53, 56), "cancel_craft.png")
    _LEVEL_UP = Button((1150, 515), (1227, 494, 45, 45), "level_up.png")

    _SEARCHBAR = (1300, 190)
    _ADDED_REGION = (40, 1020, 360, 60)
    _ITEM_REGION = (1243, 232, 562, 710)
    _UPPER_ITEM_REGION = (1240, 230, 568, 191)
    _SLOTS_REGION = (1074, 500, 60, 23)
    _REMOTE_INVENTORY = (1346, 563, 345, 43)
    _CAPPED_ICON = (1210, 230, 55, 54)

    def __init__(
        self,
        entity_name: str,
        craftables: Optional[list[Item]] = None,
        capacity: Optional[str | int] = None,
    ) -> None:
        super().__init__()
        self._name = entity_name
        self._capacity = capacity
        self._craftables = craftables
        if isinstance(capacity, str):
            self._capacity = get_filepath(capacity)

        self._contents: dict[str, int] = {}

    def __str__(self) -> str:
        return f"Inventory of {self._name} with max slots {self._capacity}"

    @property
    def contents(self) -> dict[str, int]:
        return self._contents

    @property
    def capacity(self) -> int | str | None:
        return self._capacity

    @property
    def craftables(self) -> list[Item] | None:
        return self._craftables

    @final
    def add_contents(self, item: Item, stacks: int) -> None:
        try:
            self._contents[item.name] += stacks
        except KeyError:
            self._contents[item.name] = stacks

    @final
    def set_content(self, item: Item, stacks: Optional[int] = None) -> None:
        if stacks is not None:
            self._contents[item.name] = stacks
        else:
            self._contents[item.name] = self.count(item)

    @final
    def locate_button(self, button: Button, **kwargs) -> bool:
        assert button.template is not None and button.region is not None
        return (
            self.window.locate_template(button.template, button.region, **kwargs)
            is not None
        )

    def is_open(self) -> bool:
        """Checks if the inventory is open."""
        return self.locate_button(
            self._INVENTORY_TAB, confidence=0.8
        ) or self.locate_button(self._CRAFTING_TAB, confidence=0.8)

    def open(self, default_key: bool = True, max_duration: int = 10) -> None:
        """Opens the inventory using the 'target inventory' keybind by default.

        Set `default_key` to `False` to use 'E' instead, which will allow to access
        entities in weird spots.

        Beware that using the non-default key regularly is not recommended.

        Raises:
        ----------
        `InventoryNotAccessibleError`
            If the inventory could not be opened within the max duration
        """
        attempts = 0
        while not self.is_open():
            attempts += 1

            key = self.keybinds.target_inventory if default_key else self.keybinds.use
            self.press(key)

            if await_event(self.is_open, max_duration=config.INVENTORY_OPEN_INTERVAL):
                break

            if attempts >= (
                max_duration * config.TIMER_FACTOR / config.INVENTORY_OPEN_INTERVAL
            ):
                raise InventoryNotAccessibleError(f"Failed to access {self._name}!")
        self._await_receiving_remove_inventory()

    def close(self) -> None:
        """Closes the inventory using the 'target inventory' keybind.

        Raises:
        ----------
        `InventoryNotClosableError` if the inventory did not close
        within 30 seconds, indicating a server / game crash.
        """
        attempts = 0
        while self.is_open():
            attempts += 1

            self.press(self.keybinds.target_inventory)
            if await_event(
                self.is_open, False, max_duration=config.INVENTORY_CLOSE_INTERVAL
            ):
                break

            if attempts > (40 * config.TIMER_FACTOR / config.INVENTORY_OPEN_INTERVAL):
                raise InventoryNotClosableError(f"Failed to close {self._name}!")
        self.sleep(0.3)

    @overload
    def scroll(self, way: Literal["up", "down"], *, rows: int = 1) -> None:
        ...

    @overload
    def scroll(self, way: Literal["up", "down"], *, pages: int = 0) -> None:
        ...

    @final
    def scroll(
        self, way: Literal["up", "down"], *, rows: int = 1, pages: int = 0
    ) -> None:
        """Scrolls up or down the inventory by a given amount of rows or pages.

        Parameters
        ----------
        way :class:`Literal["up", "down"]`:
            The direction to scroll, either up or down.

        rows :class:`int`:
            The number of rows to scroll, default 1

        pages :class:`int`:
            Alternative to `rows`, scroll by page rather than row. 1 page = 7 rows.
        """
        if way not in ["up", "down"]:
            raise ValueError(f'Expected one of {["up", "down"]}, got "{way}".')

        if not self.is_open():
            raise InventoryNotOpenError

        if pages:
            rows = pages * 7

        for _ in range(rows):
            self.mouse_scroll(2.91 * (1 if way == "up" else -1))

    @final
    def search(self, item: Item | str, delete_prior: bool = True) -> None:
        """Searches for an item or word in the searchbar. Then presses
        escape to tab back out of the spacebar.

        Parameters
        ----------
        items :class:`Item | str`:
            The item or term to search for
        """
        if not self.is_open():
            raise InventoryNotOpenError

        self._click_searchbar(delete_prior)
        # lowercasing the term because pyautogui has a weird gist where it will
        # actually use shift + letter to capitalize it, which opens the chat somehow
        if isinstance(item, str):
            pg.typewrite(item.lower(), interval=0.001)
        else:
            pg.typewrite(item.search_name.lower(), interval=0.001)
        # escape out of the searchbar so presing f closes the inventory
        self.press("esc")

    @final
    def drop_all(self, items: Optional[Iterable[Item | str]] = None) -> None:
        """Searches for an iterable of Items or words and drops all. If no
        items are passed, it simply drops without searching for anything.

        Parameters
        ----------
        items :class:`Iterable[Item | str]`: [Optional]
            An iterable of items to search for, then drop
        """
        if not self.is_open():
            raise InventoryNotOpenError

        if items is None:
            self.click_at(self._DROP_ALL.location)
            return

        # already know someone isnt gonna check the typehints and pass
        # an item or string on its own :D
        if not isinstance(items, Iterable):
            items = {items}
        else:
            items = set(items)

        for item in items:
            self.search(item)
            self.click_at(self._DROP_ALL.location)

    @final
    def transfer_all(
        self, items: Optional[Iterable[Item | str] | Item | str] = None
    ) -> None:
        """Searches for an iterable of Items or words and transfers all. If no
        items are passed, it simply transfers all without searching for anything.

        Parameters
        ----------
        items :class:`Iterable[Item | str]`: [Optional]
            An iterable of items to search for, then transfer
        """
        if not self.is_open():
            raise InventoryNotOpenError

        def press_button() -> None:
            while (time.time() - self.LAST_TRANSFER_ALL) < 2:
                self.sleep(0.1)

            self.click_at(self._TRANSFER_ALL.location, delay=0.2)
            Inventory.LAST_TRANSFER_ALL = time.time()

        if items is None:
            press_button()
            return

        # already know someone isnt gonna check the typehints and pass
        # an item or string on its own :D
        if isinstance(items, (str, Item)):
            items = {items}
        else:
            items = set(items)

        for item in items:
            self.search(item)
            press_button()

    @final
    def open_tab(self, tab: Literal["inventory", "crafting"]) -> None:
        """Opens the requested tab, either the inventory or crafting. Ensures
        the inventory is open before doing so to avoid punching.

        Parameters
        ----------
        tab :class:`Literal["inventory", "crafting"]`:
            The tab to open
        """
        if not self.is_open():
            raise InventoryNotOpenError

        if tab == "crafting":
            self.click_at(self._CRAFTING_TAB.location, delay=0.3)
        elif tab == "inventory":
            self.click_at(self._INVENTORY_TAB.location, delay=0.3)
        else:
            raise ValueError(f"Expected one of ['inventory', 'crafting'], got {tab}")

    @final
    def drop(self, item: Item, search: bool = True) -> None:
        """Searches for the given item and popcorns it until there is none left.

        Parameters
        ----------
        item :class:`Item`:
            The item to be popcorned
        """
        if not self.is_open():
            raise InventoryNotOpenError

        if search:
            self.search(item)

        while position := self.find(item, is_searched=search):
            self.move_to(position)
            self.press(self.keybinds.drop)

    @final
    def count(self, item: Item) -> int:
        """Returns the amount of stacks of the given item located within the inventory."""
        return len(
            self.window.locate_all_template(
                item.inventory_icon,
                region=self._ITEM_REGION,
                confidence=0.85,
                grayscale=True,
            )
        )

    @final
    def find(self, item: Item, is_searched: bool = False) -> tuple[int, int] | None:
        """Returns the position of the given item within the inventory.

        Beware that the item is grayscaled to make it compatible
        across qualities.

        Parameters:
        ------------
        item :class:`Item`:
            The item object of the item to check for
        """
        return self.window.locate_template(
            item.inventory_icon,
            region=self._UPPER_ITEM_REGION if is_searched else self._ITEM_REGION,
            confidence=0.8,
            center=True,
            grayscale=True,
        )

    @final
    def has(self, item: Item, is_searched: bool = False) -> bool:
        """Returns whether the player inventory has an item.

        Beware that the item is grayscaled to make it compatible
        across qualities.

        Parameters:
        ------------
        item :class:`Item`:
            The item object of the item to check for

        Returns whether the item is in the inventory or not.
        """
        return self.find(item, is_searched) is not None

    @overload
    def take(self, item: Item, *, amount: int) -> None:
        ...

    @overload
    def take(self, item: Item, *, stacks: int) -> None:
        ...

    @final
    def take(self, item: Item, *, stacks: int = 0, amount: int = 1) -> None:
        """Searches the given item and takes one.

        Parameters:
        -----------
        item :class:`Item`:
            The item object to search for

        stacks :class:`int`:
            The amount of stacks to take of the item

        Raises:
        -------
        `NoItemsAddedError`
            When a stack could not be transferred within 30 seconds after
            pressing 'T'

        """
        # search for the item and hit take all
        self.search(item)
        self.sleep(0.2)

        # take a specific quantity of a stack
        if not stacks:
            pos = self.find(item)
            if pos is None:
                return
            self.click_at(pos)

            for _ in range(amount):
                pg.click(clicks=2)

            return

        # take stacks only, using T-transfers
        for stack in range(stacks):
            if (pos := self.find(item)) is None:
                return

            before_take = self.count(item)
            if not stack:
                self.click_at(pos)
            self.press(self.keybinds.transfer)
            try:
                self._receive_stack(item, before_take)
            except NoItemsAddedError:
                continue

    @overload
    def manage_view_option(
        self,
        option: Literal["folder view", "show engrams", "unlearned engrams"],
    ) -> bool:
        ...

    @overload
    def manage_view_option(
        self,
        option: Literal["folder view", "show engrams", "unlearned engrams"],
        *,
        set_to: bool,
    ) -> None:
        ...

    def manage_view_option(
        self,
        option: Literal["folder view", "show engrams", "unlearned engrams"],
        *,
        set_to: Optional[bool] = None,
    ) -> bool | None:
        """Manages the inventories view option buttons. If `set_to` is not passed,
        it simply returns whether a given option is currently enabled.

        Parameters
        ----------
        option :class:`Literal["folder view", "show engrams", "unlearned engrams"]`:
            The option to check or change

        set_to :class:`Optional[bool]`:
            The state to set the option to, `None` by default
        """
        if option not in ["folder view", "show engrams", "unlearned engrams"]:
            raise ValueError(
                f'Expected one of {["folder view", "show engrams", "unlearned engrams"]}, got {option}'
            )

        if not self.is_open():
            raise InventoryNotOpenError

        buttons = {
            "folder view": self._FOLDER_VIEW,
            "show engrams": self._SHOW_ENGRAMS,
            "unlearned engrams": self._UNL_ENGRAMS,
        }
        state = self.locate_button(buttons[option], confidence=0.8)
        if set_to is None:
            return state

        if state != set_to:
            self.click_at(buttons[option].location)
        return None

    @final
    def is_capped(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/capped.png",
                region=self._CAPPED_ICON,
                confidence=0.75,
                grayscale=True,
            )
            is not None
        )

    @final
    def is_crafting(self) -> bool:
        return self.locate_button(self._STOP_CRAFTING, confidence=0.7)

    @final
    def has_level_up(self) -> bool:
        return pg.pixelMatchesColor(1161,524, (0,0,0), tolerance=3)

    def level_skill(self, skill: str, times: int) -> None:
        pos = self.LEVEL_UP_BUTTONS[skill]
        self.move_to(pos)
        
        for _ in range(times):
            pg.click()

    def stop_crafting(self) -> None:
        if not self.is_crafting():
            self.click_at(self._STOP_CRAFTING.location)

    def select_slot(self, idx: int = 0) -> None:
        """Moves to the first slot"""
        self.move_to(get_center(self.SLOTS[idx]))

    def get_folder_index(self) -> int:
        """Returns the number of the crop plot in the stack by checking for
        the folder name from AAA to HHH, being 1 to 9."""
        for _ in range(3):
            for index, option in enumerate(self._FOLDERS, start=1):
                if self.window.locate_template(
                    f"{self.PKG_DIR}/assets/interfaces/folder_{option}.png",
                    region=(1240, 290, 55, 34),
                    confidence=0.9,
                ):
                    return index
            self.sleep(0.5)
        raise UnknownFolderIndexError

    def create_folder(self, name: str) -> None:
        """Creates a folder in the inventory at the classes folder button"""
        if not self.is_open():
            raise InventoryNotOpenError

        set_clipboard(name)

        self.click_at(1585, 187)
        self.sleep(0.3)
        pg.hotkey("ctrl", "v", interval=0.2)
        self.sleep(0.3)
        self.click_at(961, 677)
        self.sleep(0.5)
        self.click("left")

    def craft(self, item: Item, amount: int) -> None:
        """Crafts the given amount of the given item. Spams 'A' if
        there is more than 50 to craft.
        """
        self.search(item)
        self.click_at(1294, 290, delay=1)

        if amount < 50:
            for _ in range(amount):
                self.press("e")
                self.sleep(0.3)
            return

        for _ in range(math.ceil(amount / 100)):
            self.press("a")
            self.sleep(0.5)

    def get_slots(self) -> int:
        """Attempts to OCR the amount of slots occupied in the structure.
        Returns the OCR'd amount as an integer, otherwise -1 on failure.
        """
        if not self.is_open():
            raise InventoryNotOpenError

        slots = self.window.grab_screen((1090, 503, 31, 15))
        masked = self.window.denoise_text(
            slots, (251, 227, 124), variance=27, dilate=False
        )
        result: str = tes.image_to_string(
            masked, config="-c tessedit_char_whitelist=0123456789lIWVSi --psm 10 -l eng"
        ).rstrip()
        if not result:
            return -1
        to_replace = {
            "/": "7",
            "l": "1",
            "I": "1",
            "W": "11",
            "V": "1",
            "i": "1",
            "S": "5",
        }
        for k, v in to_replace.items():
            result = result.replace(k, v)

        # replace mistaken "0"s, strip off newlines
        return int(result.replace("O", "0")) or -1

    def is_full(self) -> bool:
        """Checks if the vault is full, raises an `AttributeError` if no
        max slot image path was defined."""
        if self._capacity is None:
            raise AttributeError(
                f"Unabled to check slots, missing 'capacity' for '{self._name}'"
            )
        if not self.is_open():
            raise InventoryNotOpenError

        if isinstance(self._capacity, str):
            return (
                self.window.locate_template(
                    self._capacity, region=self._SLOTS_REGION, confidence=0.9
                )
                is not None
            )
        return self._capacity - 5 <= self.get_slots() <= self._capacity

    def received_item(self) -> bool:
        """Checks if an item was added by matching for the added template"""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/added.png",
                region=self._ADDED_REGION,
                confidence=0.7,
            )
            is not None
        )
    
    def delete_search(self) -> None:
        """Deletes the last term in the searchbar by selecting all of it,
        deleting it with backspace and then escaping out."""
        if not self.is_open():
            raise InventoryNotOpenError

        self._click_searchbar(delete_prior=True)
        self.press("backspace")
        self.press("esc")

    def transfer_top_row(self, speed: int | float = 0.2) -> None:
        if not self.is_open():
            raise InventoryNotOpenError

        for idx, slot in enumerate(self.SLOTS, start=1):
            pg.moveTo(get_center(slot))
            self.press(self.keybinds.transfer)
            self.sleep(speed)

            if idx >= 6:
                return

    def get_amount_transferred(
        self, item: Item, mode: Literal["rm", "add"] = "rm"
    ) -> int:
        """OCRs the amount transferred into the inventory by checking for the
        amount on the lefthand side of the screen."""
        # prepare the image
        for _ in range(10):
            roi = self._get_transferred_frame(item, mode)
            if roi:
                break
            self.sleep(0.1)

        if not roi:
            return 0

        img = self.window.grab_screen(roi, convert=False)
        img = self.window.denoise_text(
            img, denoise_rgb=(255, 255, 255), variance=10, upscale=True, upscale_by=3
        )
        # get the raw tesseract result
        raw_result: str = tes.image_to_string(
            img,
            config="-c tessedit_char_whitelist='0123456789liIxObL ' --psm 7 -l eng",
        )
        try:
            # replace all the mistaken "1"s
            for char in ["I", "l", "i", "b", "L"]:
                raw_result = raw_result.replace(char, "1")

            # replace mistaken "0"s, strip off newlines
            filtered = raw_result.replace("O", "0").replace("~", "x").rstrip()

            # find the x to slice out the actual number
            x = filtered.find("x")
            if x == -1 or not filtered or filtered == "x":
                return 0
            return int(filtered[:x])
        except:
            return 0

    def _get_transferred_frame(
        self, item: Item, mode: Literal["rm", "add"] = "rm"
    ) -> tuple[int, int, int, int] | None:
        if not item.added_icon:
            raise AttributeError("Cant find transferred frame without 'added_icon'")

        icon_pos = self.window.locate_template(
            item.added_icon, region=(0, 970, 160, 350), confidence=0.7
        )
        if icon_pos is None:
            return None

        # get our region of interest
        text_start_x = icon_pos[0] + icon_pos[2]

        if mode == "rm":
            text_end = text_start_x + self.window.convert_width(130), icon_pos[1]
        else:
            text_end = text_start_x + self.window.convert_width(95), icon_pos[1]

        roi = (
            *text_end,
            self.window.convert_width(290),
            self.window.convert_height(25),
        )

        if item.added_text is None:
            raise AttributeError(f"You did not define an 'added_text' for {item}!")

        # find the items name to crop out the numbers
        name_text = self.window.locate_template(
            item.added_text, region=roi, confidence=0.7, convert=False
        )
        if name_text is None:
            return None

        # get our region of interest (from end of "Removed:" to start of "Item")
        return (
            *text_end,
            self.window.convert_width(int(name_text[0] - text_end[0])),
            self.window.convert_height(25),
        )

    def _await_receiving_remove_inventory(self) -> None:
        """Waits until 'Receiving Remote Inventory' disappears.

        Raises:
        ------------
        `ReceivingRemoveInventoryTimeout` if it did not disappear after 30
        seconds, usually indicating a server crash.
        """
        c = 0
        while self._receiving_remote_inventory():
            self.sleep(0.1)
            c += 1
            if c > 300:
                raise ReceivingRemoveInventoryTimeout(
                    "Timed out waiting to receive remote inventory!"
                )

    def _click_searchbar(self, delete_prior: bool = True) -> None:
        """Clicks into the searchbar"""
        self.click_at(self._SEARCHBAR)
        if not delete_prior:
            return

        with pg.hold("ctrl"):
            pg.press("a")

    def _receiving_remote_inventory(self) -> bool:
        """Checks if the 'Receiving Remote Inventory' text is visible."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/remote_inventory.png",
                region=self._REMOTE_INVENTORY,
                confidence=0.8,
            )
            is not None
        )

    def _has_engram(self, item: Item) -> bool:
        """Checks if the given item can be crafted."""
        if self._craftables is None:
            raise AttributeError("Unexpected call, `craftables` is `None`.")

        if item not in self._craftables:
            raise ValueError(f"'{item.name}' is not part of '{self._name}' craftables!")

        self.search(item)
        self.sleep(0.5)
        return (
            self.window.locate_template(
                item.inventory_icon, self._ITEM_REGION, confidence=0.8
            )
            is not None
        )

    def _receive_stack(self, item: Item, before: int) -> None:
        start = time.time()

        while self.count(item) == before:
            self.sleep(0.05)
            if timedout(start, 5):
                raise NoItemsAddedError(item.name)

        print(f"Receiving stack took {time.time() - start}")



from typing import final

from ...items import FLINT, Item
from .._button import Button
from .structure import Structure


@final
class IndustrialGrinder(Structure):
    """Represents the Industrial Grinder in Ark.

    The industrial grinder extends the abilities the `Structure` class
    by adding methods to grind it's contents or specific items. For automation
    purposes the grinder plays a big role for easy polymer access.

    Beware that, while you may consider polymer, paste, ingots or pearls
    to be craftables of the `Industrial Grinder`, they are not considered as such.

    It is defined as a toggleable, but items that are a result of a grinding
    session should not be considered crafted. Chitin, flint, thatch and stone
    however are considered craftables, as they can be crafted through an engram.
    """

    GRIND_ALL = Button((969, 663), (740, 570, 444, 140), "grind_all_items.png")
    GRIND_STACK = Button((963, 701))

    def __init__(self) -> None:
        super().__init__(
            name="Industrial Grinder",
            action_wheel="assets/wheels/industrial_grinder.png",
            craftables=[FLINT],
            toggleable=True,
        )

    def can_grind(self) -> bool:
        """Checks if the grinder can grind"""
        return self.inventory.locate_button(self.GRIND_ALL, confidence=0.85)

    def grind_all(self) -> bool:
        """Presses the grind all button if it is available.
        Returns whether items got grinded or not.
        """
        if not self.can_grind():
            return False

        self.click_at(self.GRIND_ALL.location, delay=0.3)
        return True

    def grind_one(self, item: Item) -> None:
        if (pos := self.inventory.find(item)) is None:
            return

        self.click_at(pos, delay=0.3)
        self.click_at(self.GRIND_STACK.location, delay=0.3)



from ..._ark import Ark

# common tesseract mistakes to account for in any tribelog messages
# could use regex for it to make it more powerful
CONTENTS_MAPPING = {
    "{": "(",
    "}": ")",
    "[": "(",
    "]": ")",
    "((": "(",
    "))": ")",
    "Owll": "Owl)",
    "\n": " ",
    "!!": "!",
    "Lvi": "Lvl",
    "Lyi": "Lvl",
    "Gaby": "Baby",
    "destroved": "destroyed",
    "destroyedl": "destroyed!",
    "destroyedl!": "destroyed!",
    "destrayed!": "destroyed!",
    "destraved": "destroyed",
    "Giaanotosaurus": "Giganotosaurus",
    "Giganotosaurusl": "Giganotosaurus)",
    "Giganotosaurusi": "Giganotosaurus)",
    "Large Gear Trap": "Large Bear Trap",
    "C4 Charae": "C4 Charge",
    "C4 Charce": "C4 Charge",
    "C4& Charae": "C4 Charge",
    "C4 Charace": "C4 Charge",
    "C4 Chareae": "C4 Charge",
    "(Pin Coded!": "(Pin Coded)",
    "(Pin Codedl": "(Pin Coded)",
    "tExo": "(Exo",
    "”": "''",
    "“": "'",
    '"': "'",
    "‘": "'",
    "’": "'",
    "''": "'",
    "hiled": "killed",
    "hilied": "killed",
    "Dcorframe": "Doorframe",
    "Ccorframe": "Doorframe",
    "Doaorframe": "Doorframe",
    "Doarframe": "Doorframe",
    "Ceilina": "Ceiling",
    "Ficor": "Floor",
    "iTek Turret!": "(Tek Turret)",
    "iTek Turret": "(Tek Turret",
    " Tek Turreti": " (Tek Turret)",
    "(Tek Turret!": "(Tek Turret)",
    " Tek Turret)": " (Tek Turret)",
    "fCarbonemvsl": "(Carbonemys)",
    "Carbonemvs": "Carbonemys",
    "(Carbonemysl!": "(Carbonemys)!",
    "iCarbonemysl": "(Carbonemys)",
    "iCarbonemysi": "(Carbonemys)",
    "(Shadowmane!": "(Shadowmane)!",
    "Desmadus": "Desmodus",
    "Desmoadus": "Desmodus",
    "(€lonel": "[CLONE]",
    "Liahtnina": "Lightning",
    "tLightning": "(Lightning",
    "Wvvern": "Wyvern",
    "Wyvern!": "Wyvern!)",
    "fLi": "(Li",
    "ernl": "ern)",
    "Fiardhawk": "Fjordhawk",
    "fF": "(F",
    "wki": "wk)",
    "iTek": "(Tek",
    "iTribe": "(Tribe",
    "i!": ")!",
    "saurus!": "saurus)!",
    "(Fin": "(Pin",
    "Caded": "Coded",
    "Codedl": "Coded)",
    "Daeedon": "Daeodon",
    "Daeodon!": "Daeodon)",
    "Heavv": "Heavy",
    "fHeavy": "(Heavy",
    "Astradelphis": "Astrodelphis",
    "Astredelphis": "Astrodelphis",
    "fAstra": "(Astra",
    "Astrodelphisi": "Astrodelphis)",
    "Vetonasaur": "Velonasaur",
    "saurl": "saur)",
    "Preranodon": "Pteranodon",
    "Preranodon!": "Pteranodon)",
    "iDesmodus": "(Desmodus",
    "R - ": "R-",
    "R- ": "R-",
    "R -": "R-",
    "IR-": "(R-",
    "iTribe": "Tribe",
    "i(": "(",
    "!)": ")",
    "fExo": "(Exo",
    "Exo-Mete": "Exo-Mek",
    "Exo-Met": "Exo-Mek",
}

# RGB to denoise with if the templates are located in the tribelog message
DENOISE_MAPPING: dict[tuple[int, int, int], str | list] = {
    (255, 0, 0): [
        f"{Ark.PKG_DIR}/assets/tribelog/tribelog_red_your.png",
        f"{Ark.PKG_DIR}/assets/tribelog/tribelog_enemy_destroyed.png",
    ],
    (208, 3, 211): f"{Ark.PKG_DIR}/assets/tribelog/tribelog_purple_your.png",
    (158, 76, 76): f"{Ark.PKG_DIR}/assets/tribelog/tribelog_sensor.png",
}

# Denoise RGB indicating a certain tribelog event
EVENT_MAPPING: dict[tuple, str] = {
    (255, 0, 0): "Something destroyed!",
    (208, 3, 211): "Something killed!",
    (158, 76, 76): "Tek Sensor triggered!",
}

# common mistakes to replace in the daytime OCR
DAYTIME_MAPPING: dict[str, str] = {
    "|": "",
    "I": "1",
    "v": "y",
    "O": "0",
    ".": ",",
    "l": "1",
    "i": "1",
    "S": "5",
    "B": "8",
    "Dayy": "Day",
    ";": "",
    "]": "1"
}

# terms to prevent alerting for
INGORED_TERMS: list[str] = ["C4", "Baby"]


from .item import Item
from .resources import ELECTRONICS, METAL_INGOT, ORGANIC_POLYMER, PASTE, ELEMENT

AUTO_TURRET = Item(
    name="Auto Turret",
    search_name="auto turret",
    recipe={PASTE: 50, ELECTRONICS: 70, METAL_INGOT: 140, ORGANIC_POLYMER: 20},
    stack_size=1,
    inventory_icon="assets/items/structures/auto_turret.png",
)

HEAVY_AUTO_TURRET = Item(
    name="Heavy Auto Turret",
    search_name="heavy",
    stack_size=1,
    inventory_icon="assets/items/structures/heavy_auto_turret.png",
    recipe={
        PASTE: 150,
        ELECTRONICS: 200,
        METAL_INGOT: 400,
        ORGANIC_POLYMER: 50,
        AUTO_TURRET: 1,
    },
)

TEK_TURRET = Item(
    name="Tek Turret",
    search_name="tek turret",
    stack_size=1,
    inventory_icon="assets/items/structures/tek_turret.png",
    recipe={
        PASTE: 50,
        ELECTRONICS: 100,
        METAL_INGOT: 100,
        ELEMENT: 3,
        ORGANIC_POLYMER: 50,
    },

)

BEHEMOTH_GATEWAY = Item(
    name="Behemoth Gateway",
    search_name="behemoth",
    stack_size=100,
    inventory_icon="assets/items/structures/behemoth_gateway.png",
)

BEHEMOTH_GATE = Item(
    name="Behemoth Gate",
    search_name="behemoth",
    stack_size=100,
    inventory_icon="assets/items/structures/behemoth_gate.png",
)

TREE_PLATFORM = Item(
    name="Tree Platform",
    search_name="tree",
    stack_size=100,
    inventory_icon="assets/items/structures/tree_platform.png",
)

CROP_PLOT = Item(
    name="Large Crop Plot",
    search_name="crop",
    stack_size=100,
    inventory_icon="assets/items/structures/crop_plot.png",
)

RESERVOIR = Item(
    name="Metal Water Reservoir",
    search_name="reservoir",
    stack_size=100,
    inventory_icon="assets/items/structures/reservoir.png",
)

WARDRUMS = Item(
    name="Wardrums",
    search_name="wardrum",
    stack_size=100,
    inventory_icon="assets/items/structures/wardrum.png",
)

METAL_FOUNDATION = Item(
    name="Metal Foundation",
    search_name="metal foundation",
    stack_size=100,
    inventory_icon="assets/items/structures/metal_foundation.png",
    recipe={METAL_INGOT: 50, PASTE: 15},
)

METAL_TRIANGLE = Item(
    name="Metal Triangle Foundation",
    search_name="metal triangle foundation",
    stack_size=100,
    inventory_icon="assets/items/structures/metal_triangle.png",
    recipe={METAL_INGOT: 25, PASTE: 8},
)

METAL_GATEWAY = Item(
    name="Metal Gateway",
    search_name="gateway",
    stack_size=100,
    inventory_icon="assets/items/structures/metal_gateway.png",
    recipe={METAL_INGOT: 170, PASTE: 50},
)

METAL_GATE = Item(
    name="Metal Gate",
    search_name="gate",
    stack_size=100,
    inventory_icon="assets/items/structures/metal_gate.png",
    recipe={METAL_INGOT: 35, PASTE: 10},
)



import time
from pathlib import Path
from typing import Optional

import pyautogui as pg  # type: ignore[import]
import pydirectinput  # type: ignore[import]
from pynput.mouse import Button, Controller  # type: ignore[import]

from ._helpers import state_checker
from .settings import InputSettings, UserSettings
from .window import ArkWindow

pg.FAILSAFE = False
pydirectinput.FAILSAFE = False


class Ark:
    """Base parent class for all classes representing objects in ark
    Provides access to the games window, mouse and keypress simulation,
    and program state checking.
    """

    PKG_DIR = str(Path(__file__).parent)
    window: ArkWindow = None  # type:ignore[assignment]
    keybinds: InputSettings = None  # type:ignore[assignment]
    settings: UserSettings = None  # type:ignore[assignment]
    mouse = Controller()

    def __init__(self, reinit: bool = False) -> None:
        if Ark.window is None or reinit:
            Ark.window = ArkWindow()

        if Ark.keybinds is None or reinit:
            Ark.keybinds = InputSettings.load()

        if Ark.settings is None or reinit:
            Ark.settings = UserSettings.load()

    @state_checker
    def sleep(self, duration: int | float) -> None:
        """Sleeps for a given duration"""
        time.sleep(duration)

    @state_checker
    def move_to(
        self,
        x: Optional[int | tuple] = None,
        y: Optional[int | tuple] = None,
        convert: bool = True,
    ) -> None:
        """Moves to the given position, scales passed coordinates by default.
        Parameters
        -----------
        x, y: :class:`int` | `tuple`
            The coordinates to move to, normalized

        convert: :class:`bool`
            Convert the coordinate to the current resolution

        ignore_limits: :class:`bool`:
            Whether to ignore the max x and y axis thats allowed to move to
        """
        pos: tuple[int, int] = pg._normalizeXYArgs(x, y)
        x, y = pos

        # we may not need to convert if the point comes from a template match
        if convert:
            x, y = self.window.convert_point(x, y)
        pg.moveTo(x, y)

    @state_checker
    def press(self, key: str) -> None:
        """Presses the given key"""
        self.window.set_foreground()
        if key not in ["thumbmousebutton2", "thumbmousebutton"]:
            pg.press(key)
            return

        # use pynputs Controller to emulate side mouse button presses
        self.mouse.click(Button.x1 if key != "thumbmousebutton2" else Button.x2)

    @state_checker
    def mouse_scroll(self, amount: int) -> None:
        self.mouse.scroll(0, amount)

    @state_checker
    def click(self, button: str) -> None:
        """Presses the given button"""
        pg.click(button=button)

    @state_checker
    def click_with_delay(self, delay: float | int = 0.2) -> None:
        """Left clicks with a given delay."""
        self.sleep(delay)
        pg.click()
        self.sleep(delay)

    def click_at(
        self, x=None, y=None, button: str = "left", delay: float = 0.1, clicks: int = 1
    ):
        """Moves to a given location and clicks with the mouse.
        Parameters:
        ----------
        pos: :class:`tuple`
            The (x,y) coordinates of the point
        button: :class:`str`
            The button to press
        """
        self.window.set_foreground()
        x, y = pg._normalizeXYArgs(x, y)
        self.move_to(x, y)
        self.sleep(delay)
        pg.click(button=button, clicks=clicks)
        self.sleep(0.1)



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Optional

import dacite

from .. import config

@dataclass
class UserSettings:
    """Represents the GameUserSettings.ini"""

    path: Path = field(init=False)
    ui_scaling: float
    fov_multiplier: float
    left_right_sens: float
    up_down_sens: float
    hide_item_names: bool
    show_item_tooltips: bool
    auto_chatbox: bool
    toggle_hud: bool
    disable_menu_transitions: bool
    resolution_x: int
    resolution_y: int
    server_filter: int
    last_server: str
    reverse_logs: bool
    local_show_all_items: bool
    remote_show_all_items: bool
    sort_type: int
    remote_sort_type: int
    remote_show_engrams: bool
    remote_hide_unlearned_engrams: bool
    in_remote_inventory: bool

    @staticmethod
    def load(path: Optional[str] = None) -> UserSettings:
        """Loads the settings from GameUserSettings.ini, using the `ARK_PATH`
        provided in the configs or an alternatively passed path."""
        if path is None:
            path = f"{config.ARK_PATH}\ShooterGame\Saved\Config\WindowsNoEditor\GameUserSettings.ini"

        with open(path) as f:
            contents = f.readlines()

        settings: dict[str, float | bool | str | Path] = {}
        settings["path"] = Path(path)

        # keep track of the session occurrences so we can find the last joined
        # server for the selected category, which is stored as an integer from 0-7
        session_occurences = 0
        for line in contents:
            if line.startswith("[ScalabilityGroups]"):
                break

            if "=" not in line:
                continue

            if "LastJoinedSessionPerCategory" in line and not settings.get("last_server"):
                if session_occurences == settings.get("server_filter"):
                    settings["last_server"] = line.split("=")[1].strip().strip('"')
                    continue
                else:
                    session_occurences += 1

            option, value = line.rstrip().split("=")
            setting = _KEY_MAP.get(option)

            if setting is not None:
                settings[setting] = _set_type(value)  # type: ignore[assignment]

        return dacite.from_dict(UserSettings, settings)

    @property
    def last_modified(self) -> str:
        epoch = self.path.stat().st_mtime
        return datetime.fromtimestamp(epoch).strftime("%Y-%m-%d %H:%M:%S")

    def listen_for_change(self) -> None:
        """Listens to any changes made to the file, this is done by loading the
        data of the last modified timestamp, and then waiting for the timestamp
        to change. Once it has we compare the new data to the previous data and
        notify about any changes.
        
        Particularly useful to find the name of a setting in the .ini by just
        changing it ingame, listening to the changes and see what value gets
        changed."""
        last_change = self.last_modified
        with open(self.path) as f:
            old_data = f.readlines()

        while self.last_modified == last_change:
            pass

        print("Change detected in GameUserSettings.ini!")
        while True:
            try:
                with open(self.path) as f:
                    new_data = f.readlines()
            except PermissionError:
                continue
            else:
                break

        for old_line, new_line in zip(old_data, new_data):
            if old_line != new_line:
                print(f"Setting '{old_line.strip()}' changed to '{new_line.strip()}'!")


_KEY_MAP = {
    "UIScaling": "ui_scaling",
    "FOVMultiplier": "fov_multiplier",
    "LookLeftRightSensitivity": "left_right_sens",
    "LookUpDownSensitivity": "up_down_sens",
    "HideItemTextOverlay": "hide_item_names",
    "bEnableInventoryItemTooltips": "show_item_tooltips",
    "bShowChatBox": "auto_chatbox",
    "bToggleExtendedHUDInfo": "toggle_hud",
    "bDisableMenuTransitions": "disable_menu_transitions",
    "ResolutionSizeX": "resolution_x",
    "ResolutionSizeY": "resolution_y",
    "LastServerSearchType": "server_filter",
    "bReverseTribeLogOrder": "reverse_logs",
    "bLocalInventoryItemsShowAllItems": "local_show_all_items",
    "LocalItemSortType": "sort_type",
    "RemoteItemSortType": "remote_sort_type",
    "bRemoteInventoryShowEngrams": "remote_show_engrams",
    "bInventoryHideUnlearnedEngrams": "remote_hide_unlearned_engrams",
    "bRemoteInventoryCraftingShowAllItems": "remote_show_all_items",
    "bRemoteInventoryShowCraftables": "in_remote_inventory"
}


def _set_type(val: str) -> str | bool | float | int:
    """Sets the value from the .ini to the correct type"""
    if val == "True":
        return True

    elif val == "False":
        return False
    try:
        if "." in val:
            return float(val)
        return int(val)
    except ValueError:
        return val



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import dacite

from .. import config


@dataclass
class InputSettings:
    """Represents the Input.ini"""

    console: str
    crouch: str
    drop: str
    inventory: str
    prone: str
    target_inventory: str
    toggle_hud: str
    hud_info: str
    use: str
    logs: str
    transfer: str
    hotbar_0: str
    hotbar_1: str
    hotbar_2: str
    hotbar_3: str
    hotbar_4: str
    hotbar_5: str
    hotbar_6: str
    hotbar_7: str
    hotbar_8: str
    hotbar_9: str

    @staticmethod
    def load(path: Optional[str] = None) -> InputSettings:
        """Loads the settings from input.ini, using the `ARK_PATH` provided
        in the configs or an alternatively passed path."""
        if path is None:
            path = f"{config.ARK_PATH}\ShooterGame\Saved\Config\WindowsNoEditor\Input.ini"

        try:
            with open(path, encoding="utf-8") as f:
                contents = f.readlines()
        except UnicodeDecodeError:
            with open(path, encoding="utf-16") as f:
                contents = f.readlines()
           
        settings: dict[str, float | bool | str | Path] = {
            "console": "tab",
            "crouch": "c",
            "drop": "o",
            "inventory": "i",
            "prone": "x",
            "target_inventory": "f",
            "toggle_hud": "backspace",
            "hud_info": "h",
            "use": "e",
            "logs": "l",
            "transfer": "t",
            "hotbar_0": "0",
            "hotbar_1": "1",
            "hotbar_2": "2",
            "hotbar_3": "3",
            "hotbar_4": "4",
            "hotbar_5": "5",
            "hotbar_6": "6",
            "hotbar_7": "7",
            "hotbar_8": "8",
            "hotbar_9": "9",
        }

        settings["path"] = Path(path)
        for line in contents:
            if not "=" in line:
                continue

            if "ConsoleKeys" in line:
                action_name = "ConsoleKeys"
                key = line.split("=")[1].strip()
            else:
                pattern = r'ActionName="([^"]+)",Key=([^,]+)'
                matches = re.search(pattern, line)

                if matches is None:
                    continue

                action_name = matches.group(1)
                key = matches.group(2)

            action = _KEY_MAP.get(action_name)
            if key.lower() in _REPLACE and action is not None:
                settings[action] = str(_REPLACE.index(key.lower()))

            elif action is not None:
                settings[action] = key.lower()

        return dacite.from_dict(InputSettings, settings)


_KEY_MAP = {
    "ConsoleKeys": "console",
    "Crouch": "crouch",
    "Prone": "prone",
    "DropItem": "drop",
    "ShowMyInventory": "inventory",
    "AccessInventory": "target_inventory",
    "ToggleHUDHidden": "toggle_hud",
    "ShowExtendedInfo": "hud_info",
    "TransferItem": "transfer",
    "Use": "use",
    "ShowTribeManager": "logs",
    "UseItem1": "hotbar_1",
    "UseItem2": "hotbar_2",
    "UseItem3": "hotbar_3",
    "UseItem4": "hotbar_4",
    "UseItem5": "hotbar_5",
    "UseItem6": "hotbar_6",
    "UseItem7": "hotbar_7",
    "UseItem8": "hotbar_8",
    "UseItem9": "hotbar_9",
    "UseItem10": "hotbar_0",
    "TransferItem": "transfer",
}

_REPLACE = [
    "zero",
    "one",
    "two",
    "three",
    "four",
    "five",
    "six",
    "seven",
    "eight",
    "nine",
]



import inspect


class TerminatedError(Exception):
    """Raised when the bot has been termined by user or critical error"""


class WheelError(Exception):
    """Base class for all wheel exceptions"""


class WheelNotAccessibleError(WheelError):
    """Raised when a wheel accessed."""

    def __init__(self, wheel: str) -> None:
        self.wheel = wheel

    def __str__(self) -> str:
        return f"Wheel '{self.wheel}' could not be activated!"


class UnexpectedWheelError(WheelError):
    """Raised when a wheel was accessed, but not the correct one"""

    def __init__(self, expected: str, got: str) -> None:
        self.expected_wheel = expected
        self.got_wheel = got

    def __str__(self) -> str:
        return f"Unexpected wheel accessed. Expected '{self.expected_wheel}', got '{self.got_wheel}'!"


class ActionNotFoundError(WheelError):
    """Raised when a wheel action could not be found"""

    def __init__(self, action: str) -> None:
        self.action = action

    def __str__(self) -> str:
        return f"Failed to select action '{self.action}'!"


class InterfaceError(Exception):
    """Base class for all interface exceptions"""


class TimerNotVisibleError(InterfaceError):
    """Raised when the timer is not visible"""


class BedNotAccessibleError(InterfaceError):
    """Raised when the bed map could not be opened."""


class BedNotFoundError(InterfaceError):
    """Raised when the bed map could not be opened."""


class InventoryError(InterfaceError):
    """Base class for all inventory exceptions"""


class InventoryNotOpenError(InventoryError):
    """Raised when attempting an interaction within the
    inventory when it is not open"""

    def __init__(self) -> None:
        self.action = inspect.stack()[1][3]

    def __str__(self) -> str:
        return f"Attempted interaction '{self.action}' with closed inventory!"


class UnknownFolderIndexError(InventoryError):
    """Raised when a folder index is invalid or not determined"""

    def __str__(self) -> str:
        return f"Failed to find expected folder"


class InventoryNotAccessibleError(InventoryError):
    """Raised when the inventory cannot be accessed."""


class InventoryNotClosableError(InventoryError):
    """Raised when the inventory cannot be closed"""


class ReceivingRemoveInventoryTimeout(InventoryError):
    """Raised when the 'Receiving Remote Inventory' text does not disappear."""


class NoItemsAddedError(InventoryError):
    """Raised when items were not added to the inventory if expected."""

    def __init__(self, expected_item: str) -> None:
        self.item = expected_item

    def __str__(self) -> str:
        return f"Expected item {self.item} was not added!"


class NoItemsDepositedError(InventoryError):
    """Raised when the 'X items deposited.' message does not appear."""


class NoGasolineError(InventoryError):
    """Raised when a structure can not be turned on"""

    def __init__(self, structure_name) -> None:
        self.structure = structure_name

    def __str__(self) -> str:
        return f"{self.structure} is out of gasoline!"


class MissingItemErrror(InventoryError):
    """Raised when an item is missing"""

    def __init__(self, item: str) -> None:
        self.item = item

    def __str__(self) -> str:
        return f"{self.item} could not be found in the inventory!"


class PlayerError(Exception):
    """Base exception for all player errors"""


class PlayerDidntSpawnError(PlayerError):
    """Raised when the player could not spawn in"""


class PlayerDidntTravelError(PlayerError):
    """Raised when the travel screen could not be detected"""


class PlayerDiedError(PlayerError):
    """Raised when the player died"""

    def __init__(self, task: str) -> None:
        self.task = task

    def __str__(self) -> str:
        return f"Player died during task '{self.task}'!"


class LogsNotOpenedError(Exception):
    """Raised when the logs could not be opened"""


class ServerNotFoundError(Exception):
    """Raised when a server could not be found after 15 minutes."""


class DediNotInRangeError(Exception):
    """Raised when the dedi deposit text could not be detected"""


class DinoNotMountedError(Exception):
    """Raised when a dino cannot be mounted, either because it does not
    have a saddle, or because its not close enough."""



from .console import Console
from .escape_menu import EscapeMenu
from .hud_info import HUDInfo
from .inventories import Inventory, PlayerInventory
from .main_menu import MainMenu
from .session_list import SessionList
from .spawn_screen import SpawnScreen
from .structures import *
from .tribelog import TribeLog, TribeLogMessage
from .wheels import ActionWheel



from .item import Item
from .resources import PASTE, METAL_INGOT, CRYSTAL, ELECTRONICS, ORGANIC_POLYMER

ASSAULT_RIFLE = Item(
    name="Assault Rifle",
    search_name="ass",
    stack_size=1,
    inventory_icon="assets/items/weapons/assault_rifle.png",
)

PUMPGUN = Item(
    name="Pump-Action Shotgun",
    search_name="pump",
    stack_size=1,
    inventory_icon="assets/items/weapons/pumpgun.png",
)

SHOTGUN = Item(
    name="Shotgun",
    search_name="shotgun",
    stack_size=1,
    inventory_icon="assets/items/weapons/shotgun.png",
)

FABRICATED_SNIPER = Item(
    name="Fabricated Sniper Rifle",
    search_name="sniper",
    stack_size=1,
    inventory_icon="assets/items/weapons/fabricated_sniper.png",
)

FABRICATED_PISTOL = Item(
    name="Fabricated Pistol",
    search_name="pistol",
    stack_size=1,
    inventory_icon="assets/items/weapons/fabricated_pistol.png",
)

LONGNECK = Item(
    name="Longneck",
    search_name="longneck",
    stack_size=1,
    inventory_icon="assets/items/weapons/longneck.png",
)

SIMPLE_PISTOL = Item(
    name="Simple Pistol",
    search_name="simple",
    stack_size=1,
    inventory_icon="assets/items/weapons/simple_pistol.png",
)

SWORD = Item(
    name="Sword",
    search_name="sword",
    stack_size=1,
    inventory_icon="assets/items/weapons/sword.png",
)

PIKE = Item(
    name="Pike",
    search_name="pike",
    stack_size=1,
    inventory_icon="assets/items/weapons/pike.png",
)

HATCHET = Item(
    name="Metal Hatchet",
    search_name="hatchet",
    stack_size=1,
    inventory_icon="assets/items/weapons/hatchet.png",
)

PICK = Item(
    name="Metal Pick",
    search_name="pick",
    stack_size=1,
    inventory_icon="assets/items/weapons/pick.png",
)

TORCH = Item(
    name="Torch",
    search_name="torch",
    stack_size=1,
    inventory_icon="assets/items/weapons/torch.png",
)

CROSSBOW = Item(
    name="Crossbow",
    search_name="crossbow",
    stack_size=1,
    inventory_icon="assets/items/weapons/crossbow.png",
)

BOW = Item(
    name="bow",
    search_name="bow",
    stack_size=1,
    inventory_icon="assets/items/weapons/bow.png",
)

C4_DETONATOR = Item(
    name="C4 Remote Detonator",
    search_name="det",
    stack_size=1,
    inventory_icon="assets/items/weapons/c4_detonator.png",
    recipe={
        PASTE: 15,
        CRYSTAL: 10,
        ELECTRONICS: 50,
        METAL_INGOT: 10,
        ORGANIC_POLYMER: 20,
    },
)

ROCKET_LAUNCHER = Item(
    name="Rocket Launcher",
    search_name="rocket launcher",
    stack_size=1,
    inventory_icon="assets/items/weapons/rocket_launcher.png",
    recipe={
        PASTE: 60,
        METAL_INGOT: 50,
        ORGANIC_POLYMER: 80,
    },
)



__all__ = (
    "ActionWheel",
    "StryderWheel",
    "IndustrialGrinder",
    "ChemistryBench",
    "IndustrialForge",
    "TekPodWheel"
)



from .stryder_wheel import StryderWheel
from .tek_pod_wheel import TekPodWheel
from .wheel import ActionWheel



import math

import psutil  # type:ignore[import]

from .items import Item


def ark_is_running() -> bool:
    """Checks if the passed process is running"""
    return "ARK: Survival Evolved" in [
        process.name() for process in psutil.process_iter()
    ]


def close_ark() -> None:
    for process in psutil.process_iter():
        if process.name() == "ARK: Survival Evolved":
            process.kill()


def break_into_components(recipe: dict[Item, int]) -> dict[Item, int]:
    """Computes the total crafting cost of an item taking it's sub
    components into consideration.

    For example, an auto turret costs 140 Metal, 50 paste, 70 electronics...
    But since electronics are a sub component, the total cost would be
    210 Metal, 50 Paste, 210 Silica Pearls...

    The cost is extracted recursively until there are no sub components left.
    Cementing paste is not considered crafted by stone and chitin.
    """
    cost: dict[Item, int] = {}

    def extract_cost(item: Item, amount_needed: int) -> None:
        assert item.recipe is not None

        for item, amount in item.recipe.items():
            if item.recipe is None:
                cost[item] = cost.get(item, 0) + (amount * amount_needed)
            else:
                extract_cost(item, amount)

    for item, amount in recipe.items():
        if item.recipe is None:
            cost[item] = cost.get(item, 0) + amount
        elif amount:
            extract_cost(item, amount)
    return cost


def compute_crafting_plan(
    item_to_craft: Item, available_materials: dict[Item, int]
) -> tuple[int, dict[Item, int], dict[Item, int]]:
    """Computes how many of an item can be crafted given an dictionary of
    available materials. The item must have a `recipe` defined. When computing
    the total possible crafts, crafting sub-components is taken into consideation.

    For example, you may be lacking electronics to craft another item, but you have
    the pearls and the ingots to craft the missing electronics. The sub-components
    that need to be crafted are returned as a dictionary alongside the maximum
    possible amount to craft.

    Parameters
    ----------
    item_to_craft :class:`Item`:
        The item to craft, must define a `recipe`

    available_materials :class:`dict[Item, int]`:
        A dictionary storing the available materials to craft, missing materials
        will be complemented as 0.

    Returns
    -------
    `tuple[int, dict[Item, int], dict[Item, int]]`:
        The maxium number of items that can be crafted, the sub-components
        that need to be crafted and the total cost.
    """
    if item_to_craft.recipe is None:
        raise ValueError(f"{item_to_craft.name} cannot be crafted.")

    available_materials = available_materials.copy()

    cost = {item: 0 for item in item_to_craft.recipe}
    _complement_available_mats(available_materials, item_to_craft.recipe)

    craftable_right_away = _compute_craftable_instantly(
        item_to_craft, available_materials
    )

    for item, amount in item_to_craft.recipe.items():
        can_craft = round(amount * craftable_right_away)
        cost[item] += can_craft
        available_materials[item] -= can_craft

    pcost = cost.copy()
    extra_crafts, sub_components_to_craft = _can_craft(
        item_to_craft, available_materials, cost
    )

    if not extra_crafts:
        cost = pcost

    subcomp_order = list(sub_components_to_craft)
    subcomp_order.sort(key=lambda item: _get_component_depth(item))
    sub_components_to_craft = {k: sub_components_to_craft[k] for k in subcomp_order}

    return (
        math.floor(extra_crafts + craftable_right_away),
        sub_components_to_craft,
        cost,
    )


def _get_component_depth(item: Item, depth=0) -> int:

    if item.recipe is None:
        return depth
    else:
        depth += 1
    for item in item.recipe:
        depth = _get_component_depth(item, depth)
    return depth


def _compute_craftable_instantly(
    item_to_craft: Item, available_materials: dict[Item, int]
) -> int:
    if item_to_craft.recipe is None:
        raise ValueError(f"{item_to_craft.name} cannot be crafted.")

    craftable_right_away = min(
        available_materials[material] / amount
        for material, amount in item_to_craft.recipe.items()
    )
    return math.floor(craftable_right_away)


def _can_craft(
    item: Item,
    available: dict[Item, int],
    cost: dict[Item, int],
    thiscraft=None,
    amount=1,
    components_to_craft=None,
) -> tuple[int, dict[Item, int]]:
    assert item.recipe is not None

    master = components_to_craft is None
    if components_to_craft is None:
        components_to_craft = {}

    crafts = 0
    while True:
        if thiscraft is None or master:
            thiscraft = {}

        for component, amount_needed in item.recipe.items():
            if amount_needed <= available.get(component, 0):
                available[component] -= amount_needed
                thiscraft[component] = thiscraft.get(component, 0) + amount_needed
                continue

            elif component.recipe is None:
                return crafts, components_to_craft

            craftable, _ = _can_craft(
                component,
                available,
                cost,
                thiscraft,
                amount_needed,
                components_to_craft,
            )
            if not craftable or craftable != amount_needed:
                return crafts, components_to_craft
            else:
                components_to_craft[component] = (
                    components_to_craft.get(component, 0) + craftable
                )

        crafts += 1
        if master:
            for material, needed in thiscraft.items():
                cost[material] = cost.get(material, 0) + needed

        if crafts == amount and not master:
            return crafts, components_to_craft


def _complement_available_mats(available: dict, required: dict) -> None:
    available.update({item: 0 for item in set(required) - set(available)})



import functools
import math
import os
import time
from inspect import signature
from pathlib import Path
from threading import Thread
from typing import Any, Callable

import psutil  # type: ignore[import]
import win32clipboard  # type: ignore[import]

from .exceptions import TerminatedError
from .state import State


def state_checker(func: Callable):
    """Checks on the Threads state before executing"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        while State.paused:
            time.sleep(0.1)

        if not State.running:
            raise TerminatedError
        return func(*args, **kwargs)

    return wrapper

def get_filepath(filepath: str) -> str:
    """Validates the given filepath to allow to adjust files to the package
    path as well as loading files from the relative bot files."""
    if os.path.exists(filepath):
        return filepath.replace("/", "\\")

    abs_path = os.path.join(str(Path(__file__).parent), filepath)
    if not os.path.exists(abs_path):
        raise FileNotFoundError(f"Could not find {filepath} anywhere.")
    return abs_path.replace("/", "\\")


def timedout(timer: float, max_time: int | float) -> bool:
    """Checks whether the given timer has exceeded the maximum time"""
    return (time.time() - timer) > max_time


def yes_no(boolean_value: bool) -> str:
    return ["No", "Yes"][boolean_value]


def threaded(name: str):
    """Threads a function, beware that it will lose its return values"""

    def outer(func: Callable):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            thread = Thread(target=func, name=name, args=args, kwargs=kwargs)
            thread.start()

        return inner

    return outer


def await_event(
    func: Callable,
    expected_return_value: Any = True,
    max_duration: int | float = 5,
    ignore_annotation: bool = False,
) -> bool:
    """Awaits for the given function to return an expected value.
    Returns whether the function returned the value in the expected time.
    """

    @state_checker
    def sleep(s):
        time.sleep(s)
    if not func.__name__ == "<lambda>":
        log_str = f"Awaiting function '{func.__name__}' "
    else:
        log_str = f"Awaiting function '{func.__qualname__}' "
    if hasattr(func, "__self__"):
        log_str += f"of '{type(func.__self__).__name__}' "
    log_str += f"to return '{expected_return_value}' within {max_duration}s"
    # print(log_str)

    return_type = signature(func).return_annotation
    assert (
        return_type == type(expected_return_value) or ignore_annotation
    ), "Functions return type does not match expected return type."

    counter = 0
    while not func() == expected_return_value:
        counter += 1
        sleep(0.05)

        if (counter / 20) > max_duration:
            return False
    return True


def ark_is_running() -> bool:
    """Checks if the passed process is running"""
    return "ARK: Survival Evolved" in [
        process.name() for process in psutil.process_iter()
    ]


def close_ark() -> None:
    for process in psutil.process_iter():
        if process.name() == "ARK: Survival Evolved":
            process.kill()


def get_center(box: tuple[int, int, int, int]) -> tuple[int, int]:
    """Gets the center of a region (x, y, width, height)
    Parameters
    ----------
    box :class: `tuple[int, int, int, int]`:
        The box to get the center of
    Returns
    -------
    :class:`tuple[int, int]`:
        The center of the box
    """
    return int(box[0] + (box[2] / 2)), int(box[1] + (box[3] / 2))


def filter_points(points: set, minimum_distance: int) -> set:
    """Filters a set of points until there are no points left that are
    closer to each other than the given minimum distance
    Parameters
    ----------
    points :class:`set`:
        The set of points to filter
    minimum_distance :class:`int`:
        The minimum distance the points need to have to each other
    Returns
    -------
    :class:`set`:
        A set of the filtered points
    """
    filtered = set()

    while points:
        eps = points.pop()
        for point in points:
            if all(abs(c2 - c1) < minimum_distance for c2, c1 in zip(eps, point)):
                break
        else:
            filtered.add(eps)
    return filtered


def set_clipboard(text):
    """Puts the passed text into the clipboard to allow for pasting"""
    win32clipboard.OpenClipboard()
    win32clipboard.EmptyClipboard()
    win32clipboard.SetClipboardText(text, win32clipboard.CF_TEXT)
    win32clipboard.CloseClipboard()


def format_seconds(seconds: int) -> str:
    """Formats a number in seconds to a string nicely displaying it in
    different formats."""
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)

    days, hours, minutes = round(days), round(hours), round(minutes)
    if days:
        return f"{days} day{'s' if days != 1 else ''} {hours} hour{'s' if hours != 1 else ''} {minutes} minute{'s' if minutes != 1 else ''}"
    elif hours:
        return f"{hours} hour{'s' if hours != 1 else ''} {minutes} minute{'s' if minutes != 1 else ''}"
    elif minutes:
        return f"{minutes} minute{'s' if minutes != 1 else ''} {seconds} second{'s' if seconds != 1 else ''}"
    else:
        return f"{seconds} seconds"


def find_center(pixels: list[tuple[int, int]]) -> tuple[int, int]:
    """Finds the 'center of mass' given an iterable of points"""
    x_total = 0
    y_total = 0

    # sum up all the x and y points
    for x, y in pixels:
        x_total += x
        y_total += y

    # get the "average" point
    center_x = round(x_total / len(pixels))
    center_y = round(y_total / len(pixels))
    return center_x, center_y


def find_closest_pixel(
    pixels: list[tuple[int, int]], center: tuple[int, int]
) -> tuple[int, int]:
    """Finds the pixel closest to the center"""
    closest_pixel = pixels[0]
    closest_distance = float("inf")

    for pixel in pixels:
        distance = math.sqrt((pixel[0] - center[0]) ** 2 + (pixel[1] - center[1]) ** 2)
        if distance < closest_distance:
            # set new closest distance
            closest_pixel = pixel
            closest_distance = distance
    return closest_pixel
